from functools import partial
from typing import Optional
import torch
from torch import Tensor, nn
from torch.nn.modules.loss import _Loss

from .functional import focal_loss_with_logits, softmax_focal_loss_with_logits
from pytorch_toolbelt.utils import pytorch_toolbelt_deprecated

__all__ = ["CrossEntropyFocalLoss", "BinaryFocalLoss", "FocalLoss"]


class BinaryFocalLoss(nn.Module):
    __constants__ = ["alpha", "gamma", "reduction", "ignore_index", "normalized", "reduced_threshold", "activation"]

    def __init__(
        self,
        alpha: Optional[float] = None,
        gamma: float = 2.0,
        ignore_index: Optional[int] = None,
        reduction: str = "mean",
        normalized: bool = False,
        reduced_threshold: Optional[float] = None,
        activation: str = "sigmoid",
        softmax_dim: Optional[int] = None,
        class_weights: Optional[Tensor] = None,
    ):
        """

        :param alpha: Prior probability of having positive value in target.
        :param gamma: Power factor for dampening weight (focal strength).
        :param ignore_index: If not None, targets may contain values to be ignored.
        Target values equal to ignore_index will be ignored from loss computation.
        :param reduced: Switch to reduced focal loss. Note, when using this mode you should use `reduction="sum"`.
        :param activation: Either `sigmoid` or `softmax`. If `softmax` is used, `softmax_dim` must be also specified.

        """
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.ignore_index = ignore_index
        self.reduction = reduction
        self.normalized = normalized
        self.reduced_threshold = reduced_threshold
        self.activation = activation
        self.softmax_dim = softmax_dim

        if class_weights is not None and not torch.is_tensor(class_weights):
            class_weights = torch.tensor(list(class_weights), dtype=torch.float32)
        self.register_buffer("class_weights", class_weights, persistent=False)

        self.focal_loss_fn = partial(
            focal_loss_with_logits,
            alpha=alpha,
            gamma=gamma,
            reduced_threshold=reduced_threshold,
            reduction=reduction,
            normalized=normalized,
            ignore_index=ignore_index,
            activation=activation,
            softmax_dim=softmax_dim,
        )

        self.get_one_hot_targets = (
            self._one_hot_targets_with_ignore if ignore_index is not None else self._one_hot_targets
        )

    def __repr__(self):
        repr = f"{self.__class__.__name__}(alpha={self.alpha}, gamma={self.gamma}, "
        repr += f"ignore_index={self.ignore_index}, reduction={self.reduction}, normalized={self.normalized}, "
        repr += f"reduced_threshold={self.reduced_threshold}, activation={self.activation}, "
        repr += f"softmax_dim={self.softmax_dim},"
        repr += f"class_weights={self.class_weights.tolist()}, "
        repr += ")"
        return repr

    def forward(self, inputs: Tensor, targets: Tensor) -> Tensor:
        """
        Compute focal loss for binary classification problem.
        Args:
            inputs: [B,C,H,W]
            targets: [B,C,H,W] one-hot or [B,H,W] long tensor that will be one-hot encoded (w.r.t to ignore_index)

        Returns:

        """

        if len(targets.shape) + 1 == len(inputs.shape):
            targets = self.get_one_hot_targets(targets, num_classes=inputs.size(1))

        loss = self.focal_loss_fn(inputs, targets, class_weights=self.class_weights)
        return loss

    def _one_hot_targets(self, targets, num_classes):
        targets_one_hot = torch.nn.functional.one_hot(targets, num_classes=num_classes)
        targets_one_hot = torch.moveaxis(targets_one_hot, -1, 1)
        return targets_one_hot

    def _one_hot_targets_with_ignore(self, targets, num_classes):
        ignored_mask = targets.eq(self.ignore_index)
        targets_masked = torch.masked_fill(targets, ignored_mask, 0)
        targets_one_hot = torch.nn.functional.one_hot(targets_masked, num_classes=num_classes)
        targets_one_hot = torch.moveaxis(targets_one_hot, -1, 1)
        targets_one_hot.masked_fill_(ignored_mask.unsqueeze(1), self.ignore_index)
        return targets_one_hot


class CrossEntropyFocalLoss(nn.Module):
    """
    Focal loss for multi-class problem. It uses softmax to compute focal term instead of sigmoid as in
    original paper. This loss expects target labes to have one dimension less (like in nn.CrossEntropyLoss).

    """

    def __init__(
        self,
        gamma: float = 2.0,
        reduction: str = "mean",
        normalized: bool = False,
        reduced_threshold: Optional[float] = None,
        ignore_index: int = -100,
        class_weights: Optional[Tensor] = None,
    ):
        """

        :param alpha:
        :param gamma:
        :param ignore_index: If not None, targets with given index are ignored
        :param reduced_threshold: A threshold factor for computing reduced focal loss
        :param class_weights: A tensor of shape [C] where C is the number of classes.
               It represents weights for each class. If None, all classes are treated equally.
               Unreduced loss is multiplied by class_weights before reduction.
        """
        super().__init__()
        self.gamma = gamma
        self.reduction = reduction
        self.reduced_threshold = reduced_threshold
        self.normalized = normalized
        self.ignore_index = ignore_index
        self.register_buffer("class_weights", class_weights, persistent=False)

    def forward(self, inputs: Tensor, targets: Tensor) -> Tensor:
        """

        Args:
            inputs: [B,C,H,W] tensor
            targets: [B,H,W] tensor

        Returns:

        """
        return softmax_focal_loss_with_logits(
            inputs,
            targets,
            gamma=self.gamma,
            reduction=self.reduction,
            normalized=self.normalized,
            reduced_threshold=self.reduced_threshold,
            ignore_index=self.ignore_index,
            class_weights=self.class_weights,
        )


@pytorch_toolbelt_deprecated("Class FocalLoss is deprecated. Please use CrossEntropyFocalLoss instead.")
def FocalLoss(*input, **kwargs):
    return CrossEntropyFocalLoss(*input, **kwargs)



from .common import GenericTimmEncoder
from ..common import EncoderModule, _take, make_n_channel_input
from ...activations import ACT_RELU, get_activation_block
from pytorch_toolbelt.modules.interfaces import FeatureMapsSpecification

__all__ = ["HRNetW18Encoder", "HRNetW32Encoder", "HRNetW48Encoder", "TimmHRNetW18SmallV2Encoder"]


class HRNetTimmEncoder(GenericTimmEncoder):
    def __init__(self, encoder, first_conv_stride_one, layers):
        if first_conv_stride_one:
            encoder.conv1.stride = (1, 1)

        super().__init__(encoder, layers)
        if first_conv_stride_one:
            self.output_spec = FeatureMapsSpecification(
                channels=self.output_spec.channels, strides=tuple([s // 2 for s in self.output_spec.strides])
            )

    def forward(self, x):
        y = self.encoder.forward(x)
        return _take(y, self._layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.encoder.conv1 = make_n_channel_input(self.encoder.conv1, input_channels, mode, **kwargs)
        return self


class HRNetW18Encoder(HRNetTimmEncoder):
    def __init__(
        self, pretrained=True, use_incre_features: bool = True, layers=None, first_conv_stride_one: bool = False
    ):
        from timm.models import hrnet

        encoder = hrnet.hrnet_w18(
            pretrained=pretrained,
            feature_location="incre" if use_incre_features else "",
            features_only=True,
            out_indices=(0, 1, 2, 3, 4),
        )
        super().__init__(encoder, first_conv_stride_one=first_conv_stride_one, layers=layers)


class HRNetW32Encoder(HRNetTimmEncoder):
    def __init__(
        elf, pretrained=True, use_incre_features: bool = True, layers=None, first_conv_stride_one: bool = False
    ):
        from timm.models import hrnet

        encoder = hrnet.hrnet_w32(
            pretrained=pretrained,
            feature_location="incre" if use_incre_features else "",
            features_only=True,
            out_indices=(0, 1, 2, 3, 4),
        )
        super().__init__(encoder, first_conv_stride_one=first_conv_stride_one, layers=layers)


class HRNetW48Encoder(HRNetTimmEncoder):
    def __init__(
        elf, pretrained=True, use_incre_features: bool = True, layers=None, first_conv_stride_one: bool = False
    ):
        from timm.models import hrnet

        encoder = hrnet.hrnet_w48(
            pretrained=pretrained,
            feature_location="incre" if use_incre_features else "",
            features_only=True,
            out_indices=(0, 1, 2, 3, 4),
        )
        super().__init__(encoder, first_conv_stride_one=first_conv_stride_one, layers=layers)


class TimmHRNetW18SmallV2Encoder(HRNetTimmEncoder):
    def __init__(
        self, pretrained=True, use_incre_features: bool = True, layers=None, first_conv_stride_one: bool = False
    ):
        from timm.models import hrnet

        encoder = hrnet.hrnet_w18_small_v2(
            pretrained=pretrained,
            feature_location="incre" if use_incre_features else "",
            features_only=True,
            out_indices=(0, 1, 2, 3, 4),
        )
        super().__init__(encoder, first_conv_stride_one=first_conv_stride_one, layers=layers)



from .common import GenericTimmEncoder


class MaxVitEncoder(GenericTimmEncoder):
    def __init__(self, model_name: str, pretrained=True, **kwargs):
        super().__init__(model_name, pretrained=pretrained, **kwargs)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        from pytorch_toolbelt.modules import make_n_channel_input

        self.encoder.stem.conv1 = make_n_channel_input(self.encoder.stem.conv1, input_channels)
        return self



import numpy as np

__all__ = ["rle_decode", "rle_encode", "rle_to_string"]


def rle_encode(mask: np.ndarray):
    """
    Convert mask to EncodedPixels in run-length encoding
    from https://www.kaggle.com/stainsby/fast-tested-rle-and-input-routines
    """
    pixels = mask.T.flatten()
    # We need to allow for cases where there is a '1' at either end of the sequence.
    # We do this by padding with a zero at each end when needed.
    use_padding = False
    if pixels[0] or pixels[-1]:
        use_padding = True
        pixel_padded = np.zeros([len(pixels) + 2], dtype=pixels.dtype)
        pixel_padded[1:-1] = pixels
        pixels = pixel_padded
    rle = np.where(pixels[1:] != pixels[:-1])[0] + 2
    if use_padding:
        rle = rle - 1
    rle[1::2] = rle[1::2] - rle[:-1:2]
    return rle


def rle_to_string(runs) -> str:
    return " ".join(str(x) for x in runs)


def rle_decode(rle_str, shape, dtype) -> np.ndarray:
    s = rle_str.split()
    starts, lengths = [np.asarray(x, dtype=int) for x in (s[0:][::2], s[1:][::2])]
    starts -= 1
    ends = starts + lengths
    mask = np.zeros(np.prod(shape), dtype=dtype)
    for lo, hi in zip(starts, ends):
        mask[lo:hi] = 1
    return mask.reshape(shape[::-1]).T



from functools import partial

import cv2
import numpy as np

from ..utils import image_to_tensor

__all__ = ["mask_to_bce_target", "mask_to_ce_target", "read_binary_mask", "compute_weight_mask"]


def mask_to_bce_target(mask):
    return image_to_tensor(mask, dummy_channels_dim=True).float()


def mask_to_ce_target(mask):
    return image_to_tensor(mask, dummy_channels_dim=False).long()


def compute_weight_mask(mask: np.ndarray, edge_weight=4) -> np.ndarray:
    from skimage.morphology import binary_dilation, binary_erosion

    binary_mask = mask > 0
    weight_mask = np.ones(mask.shape[:2]).astype(np.float32)

    if binary_mask.any():
        dilated = binary_dilation(binary_mask, structure=np.ones((5, 5), dtype=np.bool))
        eroded = binary_erosion(binary_mask, structure=np.ones((5, 5), dtype=np.bool))

        a = dilated & ~binary_mask
        b = binary_mask & ~eroded

        weight_mask = (a | b).astype(np.float32) * edge_weight + 1
        weight_mask = cv2.GaussianBlur(weight_mask, ksize=(5, 5), sigmaX=5)
    return weight_mask


def _block_reduce_dominant_label(x: np.ndarray, axis):
    try:
        # minlength is +1 to num classes because we must account for IGNORE_LABEL
        minlength = np.max(x) + 1
        bincount_fn = partial(np.bincount, minlength=minlength)
        counts = np.apply_along_axis(bincount_fn, -1, x.reshape((x.shape[0], x.shape[1], -1)))
        reduced = np.argmax(counts, axis=-1)
        return reduced
    except Exception as e:
        print(e)
        print("shape", x.shape, "axis", axis)


def read_binary_mask(mask_fname: str) -> np.ndarray:
    """
    Read image as binary mask, all non-zero values are treated as positive labels and converted to 1
    Args:
        mask_fname: Image with mask

    Returns:
        Numpy array with {0,1} values
    """
    mask = cv2.imread(mask_fname, cv2.IMREAD_GRAYSCALE)
    if mask is None:
        raise FileNotFoundError(f"Cannot find {mask_fname}")

    cv2.threshold(mask, thresh=0, maxval=1, type=cv2.THRESH_BINARY, dst=mask)
    return mask



"""Common functions to marshal data to/from PyTorch

"""

import collections
import dataclasses
import functools
import warnings
import logging
from typing import Optional, Sequence, Union, Dict, List, Any, Iterable, Callable, Tuple, Mapping

import numpy as np
import torch
from torch import nn, Tensor
from torch.utils.data import Dataset, ConcatDataset
from torch.utils.data.dataloader import default_collate

from .support import pytorch_toolbelt_deprecated

__all__ = [
    "argmax_over_dim_0",
    "argmax_over_dim_1",
    "argmax_over_dim_2",
    "argmax_over_dim_3",
    "count_parameters",
    "image_to_tensor",
    "int_to_string_human_friendly",
    "logit",
    "mask_from_tensor",
    "maybe_cuda",
    "resize_as",
    "resize_like",
    "rgb_image_from_tensor",
    "sigmoid_with_threshold",
    "softmax_over_dim_0",
    "softmax_over_dim_1",
    "softmax_over_dim_2",
    "softmax_over_dim_3",
    "tensor_from_mask_image",
    "tensor_from_rgb_image",
    "to_numpy",
    "to_tensor",
    "transfer_weights",
    "move_to_device",
    "move_to_device_non_blocking",
    "describe_outputs",
    "get_collate_for_dataset",
    "get_non_wrapped_model",
    "container_to_tensor",
]


def softmax_over_dim_0(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=0)


def softmax_over_dim_1(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=1)


def softmax_over_dim_2(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=2)


def softmax_over_dim_3(x: Tensor) -> Tensor:
    return torch.softmax(x, dim=3)


def argmax_over_dim_0(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=0)


def argmax_over_dim_1(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=1)


def argmax_over_dim_2(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=2)


def argmax_over_dim_3(x: Tensor) -> Tensor:
    return torch.argmax(x, dim=3)


def sigmoid_with_threshold(x: Tensor, threshold):
    return x.float().sigmoid().gt(threshold)


def logit(x: torch.Tensor, eps=1e-5) -> torch.Tensor:
    """
    Compute inverse of sigmoid of the input.
    Note: This function has not been tested for numerical stability.
    :param x:
    :param eps:
    :return:
    """
    x = torch.clamp(x, eps, 1.0 - eps)
    return torch.log(x / (1.0 - x))


def count_parameters(
    model: nn.Module, keys: Optional[Sequence[str]] = None, human_friendly: bool = False
) -> Dict[str, int]:
    """
    Count number of total and trainable parameters of a model
    :param model: A model
    :param keys: Optional list of top-level blocks
    :param human_friendly: If True, outputs human-friendly number of paramters: 13.3M, 124K
    :return: Tuple (total, trainable)
    """
    if keys is None:
        keys = [key for key, child in model.named_children()]
    total = int(sum(p.numel() for p in model.parameters()))
    trainable = int(sum(p.numel() for p in model.parameters() if p.requires_grad))
    parameters = {"total": total, "trainable": trainable}

    for key in keys:
        try:
            if hasattr(model, key) and model.__getattr__(key) is not None:
                parameters[key] = int(sum(p.numel() for p in model.__getattr__(key).parameters()))
        except AttributeError:
            pass

    if human_friendly:
        for key in parameters.keys():
            parameters[key] = int_to_string_human_friendly(parameters[key])
    return parameters


def int_to_string_human_friendly(value: int) -> str:
    if value < 1000:
        return str(value)
    if value < 1000000:
        return f"{value / 1000.:.2f}K"
    if value < 10000000:
        return f"{value / 1000000.:.2f}M"
    if value < 100000000:
        return f"{value / 1000000.:.1f}M"
    if value < 1000000000:
        return f"{value / 1000000.:.1f}M"
    return f"{value / 1000000000.:.2f}B"


def to_numpy(x: Union[torch.Tensor, np.ndarray, Any, None]) -> Union[np.ndarray, None]:
    """
    Convert whatever to numpy array. None value returned as is.

    Args:
        :param x: List, tuple, PyTorch tensor or numpy array

    Returns:
        :return: Numpy array
    """
    if x is None:
        return None
    elif torch.is_tensor(x) and x.dtype == torch.bfloat16:
        return x.data.float().cpu().numpy()
    elif torch.is_tensor(x):
        return x.data.cpu().numpy()
    elif isinstance(x, np.ndarray):
        return x
    elif isinstance(x, (Iterable, int, float)):
        return np.array(x)
    else:
        raise ValueError("Unsupported type")


def to_tensor(x, dtype=None) -> torch.Tensor:
    if isinstance(x, torch.Tensor):
        if dtype is not None:
            x = x.type(dtype)
        return x
    if isinstance(x, np.ndarray) and x.dtype.kind not in {"O", "M", "U", "S"}:
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x
    if isinstance(x, (list, tuple)):
        x = np.ndarray(x)
        x = torch.from_numpy(x)
        if dtype is not None:
            x = x.type(dtype)
        return x

    raise ValueError("Unsupported input type" + str(type(x)))


def container_to_tensor(value: Union[np.ndarray, List, Tuple, Mapping, Any]):
    if torch.is_tensor(value):
        return value
    if isinstance(value, np.ndarray) and value.dtype.kind not in {"O", "M", "U", "S"}:
        return torch.from_numpy(value)
    if isinstance(value, list):
        return [container_to_tensor(item) for item in value]
    if isinstance(value, tuple):
        return tuple(container_to_tensor(item) for item in value)
    if isinstance(value, Mapping):
        cls = type(value)
        return cls((k, container_to_tensor(v)) for k, v in value.items())

    raise ValueError(f"Unsupported container type {type(value)}")


def image_to_tensor(image: np.ndarray, dummy_channels_dim=True) -> torch.Tensor:
    """
    Convert numpy image (RGB, BGR, Grayscale, SAR, Mask image, etc.) to tensor

    Args:
        image: A numpy array of [H,W,C] shape
        dummy_channels_dim: If True, and image has [H,W] shape adds dummy channel, so that
            output tensor has shape [1, H, W]

    See also:
        rgb_image_from_tensor - To convert tensor image back to RGB with denormalization
        mask_from_tensor

    Returns:
        Torch tensor of [C,H,W] or [H,W] shape (dummy_channels_dim=False).
    """
    if len(image.shape) not in {2, 3}:
        raise ValueError(f"Image must have shape [H,W] or [H,W,C]. Got image with shape {image.shape}")

    if len(image.shape) == 2:
        if dummy_channels_dim:
            image = np.expand_dims(image, 0)
    else:
        # HWC -> CHW
        image = np.moveaxis(image, -1, 0)
    image = np.require(image, requirements="C")
    image = torch.from_numpy(image)
    return image


@pytorch_toolbelt_deprecated("This function is deprecated, please use image_to_tensor instead")
def tensor_from_rgb_image(image: np.ndarray) -> torch.Tensor:
    return image_to_tensor(image)


@pytorch_toolbelt_deprecated("This function is deprecated, please use image_to_tensor instead")
def tensor_from_mask_image(mask: np.ndarray) -> torch.Tensor:
    return image_to_tensor(mask)


def rgb_image_from_tensor(
    image: torch.Tensor,
    mean=(0.485, 0.456, 0.406),
    std=(0.229, 0.224, 0.225),
    min_pixel_value=0.0,
    max_pixel_value=255.0,
    dtype=np.uint8,
) -> np.ndarray:
    """
    Convert numpy image (RGB, BGR, Grayscale, SAR, Mask image, etc.) to tensor

    Args:
        image: A torch tensor of [C,H,W] shape
    """
    image = np.moveaxis(to_numpy(image), 0, -1)
    mean = to_numpy(mean)
    std = to_numpy(std)
    rgb_image = max_pixel_value * (image * std + mean)
    rgb_image = np.clip(rgb_image, a_min=min_pixel_value, a_max=max_pixel_value)
    return rgb_image.astype(dtype)


def mask_from_tensor(mask: torch.Tensor, squeeze_single_channel: bool = False, dtype=None) -> np.ndarray:
    mask_np = np.moveaxis(to_numpy(mask), 0, -1)
    if squeeze_single_channel and mask_np.shape[-1] == 1:
        mask_np = np.squeeze(mask_np, -1)

    if dtype is not None:
        mask_np = mask_np.astype(dtype)
    return mask_np


def maybe_cuda(x: Union[torch.Tensor, nn.Module]) -> Union[torch.Tensor, nn.Module]:
    """
    Move input Tensor or Module to CUDA device if CUDA is available.
    :param x:
    :return:
    """
    if torch.cuda.is_available():
        return x.cuda()
    return x


logger = logging.getLogger("pytorch_toolbelt.utils")


def transfer_weights(model: nn.Module, model_state_dict: collections.OrderedDict, incompatible_shape_action="skip"):
    """
    Copy weights from state dict to model, skipping layers that are incompatible.
    This method is helpful if you are doing some model surgery and want to load
    part of the model weights into different model.
    :param model: Model to load weights into
    :param model_state_dict: Model state dict to load weights from
    :param incompatible_shape_action: What to do if shape of weight tensor is incompatible.
    Possible values are:
        - "skip" - Skip loading this tensor
        - "match_mean_std" - Initialize tensor with random values with same mean and std as source tensor
    :return: None
    """
    existing_model_state_dict = model.state_dict()

    loaded_layers = 0

    for name, value in model_state_dict.items():
        if name not in existing_model_state_dict:
            logger.debug(
                f"transfer_weights skipped loading weights for key {name}, because it does not exist in model"
            )
            continue

        existing_value = existing_model_state_dict[name]
        if value.shape != existing_value.shape:
            if incompatible_shape_action == "skip":
                logger.debug(
                    f"transfer_weights skipped loading weights for key {name}, because of checkpoint has shape {value.shape} and model has shape {existing_model_state_dict[name].shape}"
                )
                continue
            elif incompatible_shape_action == "match_mean_std":
                logger.debug(
                    f"transfer_weights found that {name} weights tensor have incompatible shape {value.shape} and model has shape {existing_value.shape}. "
                    f"Initializing with random values with same mean {existing_value.mean()} and std {existing_value.std()} from corresponding checkpoint weights tensor."
                )
                torch.nn.init.normal_(existing_value, mean=value.mean(), std=value.std())
                value = existing_value
            else:
                raise ValueError(f"Unsupported incompatible_shape_action={incompatible_shape_action}")

        try:
            model.load_state_dict(collections.OrderedDict([(name, value)]), strict=False)
            loaded_layers += 1
        except Exception as e:
            logger.debug(f"transfer_weights skipped loading weights for key {name}, because of error: {e}")

    percentage_of_layers_from_checkpoint = loaded_layers / len(model_state_dict) * 100
    percentage_of_layers_in_model = loaded_layers / len(existing_model_state_dict) * 100
    logger.info(
        f"Transferred {percentage_of_layers_from_checkpoint:.2f}% of layers from checkpoint to model, filling {percentage_of_layers_in_model:.2f}% of model layers"
    )


def resize_like(x: Tensor, target: Tensor, mode: str = "bilinear", align_corners: Union[bool, None] = True) -> Tensor:
    """
    Resize input tensor to have the same spatial dimensions as target.

    Args:
        x: Input tensor of [B,C,H,W]
        target: [Bt,Ct,Ht,Wt]
        mode:
        align_corners:

    Returns:
        Resized tensor [B,C,Ht,Wt]
    """
    return torch.nn.functional.interpolate(x, target.size()[2:], mode=mode, align_corners=align_corners)


def move_to_device_non_blocking(x: Tensor, device: torch.device) -> Tensor:
    return move_to_device(x, device, non_blocking=True)


def move_to_device(
    x: Union[Tensor, List[Tensor], Tuple[Tensor, ...], Dict[Any, Tensor]], device: torch.device, non_blocking=False
) -> Tensor:
    if torch.is_tensor(x):
        return x.to(device=device, non_blocking=non_blocking)
    elif isinstance(x, tuple):
        return tuple(move_to_device(item, device, non_blocking) for item in x)
    elif isinstance(x, list):
        return [move_to_device(item, device, non_blocking) for item in x]
    elif isinstance(x, Mapping):
        cls = type(x)
        return cls((key, move_to_device(item, device, non_blocking)) for key, item in x.items())
    return x


resize_as = resize_like


def describe_outputs(outputs: Union[Tensor, Dict[str, Tensor], Iterable[Tensor]]) -> Union[List[Dict], Dict[str, Any]]:
    """
    Describe outputs and return shape, mean & std for each tensor in list or dict (Supports nested tensors)

    Args:
        outputs: Input (Usually model outputs)
    Returns:
        Same structure but each item represents tensor shape, mean & std
    """
    if torch.is_tensor(outputs):
        if torch.is_floating_point(outputs):
            desc = dict(
                size=tuple(outputs.size()), mean=outputs.mean().item(), std=outputs.std().item(), dtype=outputs.dtype
            )
        else:
            desc = dict(size=tuple(outputs.size()), num_unique=len(torch.unique(outputs)), dtype=outputs.dtype)
    elif isinstance(outputs, collections.abc.Mapping):
        desc = {}
        for key, value in outputs.items():
            desc[key] = describe_outputs(value)
    elif dataclasses.is_dataclass(outputs):
        desc = dataclasses.asdict(outputs)
        for key, value in desc.items():
            desc[key] = describe_outputs(value)
    elif isinstance(outputs, collections.abc.Iterable):
        desc = []
        for index, output in enumerate(outputs):
            desc.append(describe_outputs(output))
    else:
        warnings.warn(f"describe_outputs is not implemeted for type {type(outputs)}")
        return str(outputs)
    return desc


def get_collate_for_dataset(
    dataset: Union[Dataset, ConcatDataset], ensure_collate_fn_are_the_same: bool = True
) -> Callable:
    """
    Return collate_fn function for dataset. By default, default_collate returned.
    If the dataset has method get_collate_fn() we will use it's return value instead.
    If the dataset is ConcatDataset, we will check whether all get_collate_fn() returns
    the same function.

    Args:
        dataset: Input dataset

    Returns:
        Collate function to put into DataLoader
    """
    collate_fn = default_collate

    if hasattr(dataset, "get_collate_fn"):
        return dataset.get_collate_fn()
    elif isinstance(dataset, ConcatDataset):
        collate_fns = [get_collate_for_dataset(ds) for ds in dataset.datasets]
        collate_fn = collate_fns[0]

        if ensure_collate_fn_are_the_same:
            for other_collate_fn in collate_fns[1:]:
                if type(other_collate_fn) != type(collate_fn):
                    raise ValueError(
                        f"Detected ConcatDataset consist of datasets with different collate functions: {type(collate_fn)} and {type(other_collate_fn)}."
                    )

                if isinstance(collate_fn, functools.partial):
                    if not _partial_functions_equal(collate_fn, other_collate_fn):
                        raise ValueError(
                            f"Detected ConcatDataset consist of datasets with different collate functions: {collate_fn} and {type(other_collate_fn)}."
                        )
                elif collate_fn != other_collate_fn:
                    raise ValueError(
                        f"Detected ConcatDataset consist of datasets with different collate functions: {collate_fn} and {other_collate_fn}."
                    )

        collate_fn = collate_fns[0]

    return collate_fn


def _partial_functions_equal(func1, func2):
    if not (isinstance(func1, functools.partial) and isinstance(func2, functools.partial)):
        return False
    are_equal = all([getattr(func1, attr) == getattr(func2, attr) for attr in ["func", "args", "keywords"]])
    return are_equal


def get_non_wrapped_model(model: nn.Module) -> nn.Module:
    """
    Return real model from (maybe) wrapped in DP / DDP

    Args:
        model:

    Returns:

    """
    from torch.nn import DataParallel
    from torch.nn.parallel import DistributedDataParallel

    if not isinstance(model, nn.Module):
        raise RuntimeError("Input model must be a subclass of nn.Module.")

    if isinstance(model, (DataParallel, DistributedDataParallel)):
        model = model.module

    return model



import numpy as np
import pytest
import torch
from pytorch_toolbelt.inference.functional import (
    unpad_xyxy_bboxes,
    pad_image_tensor,
    unpad_image_tensor,
    pad_tensor_to_size,
)
from pytorch_toolbelt.modules.encoders import make_n_channel_input
from pytorch_toolbelt.utils import match_bboxes, match_bboxes_hungarian, get_collate_for_dataset
from torch import nn
from torch.utils.data import ConcatDataset
from functools import partial


def test_pad_tensor_to_size_1d():
    x = torch.randn((1, 3, 32))

    x_padded, unpad = pad_tensor_to_size(x, (48,))

    assert x_padded.size(2) == 48
    print(x_padded[unpad].size())
    assert (x_padded[unpad] == x).all()


def test_pad_tensor_to_size_2d():
    x = torch.randn((1, 3, 32, 33))

    x_padded, unpad = pad_tensor_to_size(x, (48, 46))

    assert x_padded.size(2) == 48
    assert x_padded.size(3) == 46
    print(x_padded[unpad].size())
    assert (x_padded[unpad] == x).all()


def test_pad_tensor_to_size_3d():
    x = torch.randn((1, 3, 32, 33, 34))

    x_padded, unpad = pad_tensor_to_size(x, (48, 46, 49))

    assert x_padded.size(2) == 48
    assert x_padded.size(3) == 46
    assert x_padded.size(4) == 49
    print(x_padded[unpad].size())
    assert (x_padded[unpad] == x).all()


def test_unpad_xyxy_bboxes():
    bboxes1 = torch.rand((1, 32, 4))
    bboxes2 = torch.rand((1, 32, 4, 20))

    pad = [2, 3, 4, 5]
    bboxes1_unpad = unpad_xyxy_bboxes(bboxes1, pad, dim=-1)
    assert bboxes1_unpad.size(0) == 1
    assert bboxes1_unpad.size(1) == 32
    assert bboxes1_unpad.size(2) == 4

    bboxes2_unpad = unpad_xyxy_bboxes(bboxes2, pad, dim=2)
    assert bboxes2_unpad.size(0) == 1
    assert bboxes2_unpad.size(1) == 32
    assert bboxes2_unpad.size(2) == 4
    assert bboxes2_unpad.size(3) == 20


def test_make_n_channel_input():
    conv = nn.Conv2d(3, 16, kernel_size=3, padding=1)

    conv6 = make_n_channel_input(conv, in_channels=6)
    assert conv6.weight.size(0) == conv.weight.size(0)
    assert conv6.weight.size(1) == 6
    assert conv6.weight.size(2) == conv.weight.size(2)
    assert conv6.weight.size(3) == conv.weight.size(3)

    conv5 = make_n_channel_input(conv, in_channels=5)
    assert conv5.weight.size(0) == conv.weight.size(0)
    assert conv5.weight.size(1) == 5
    assert conv5.weight.size(2) == conv.weight.size(2)
    assert conv5.weight.size(3) == conv.weight.size(3)


@pytest.mark.parametrize(
    ["shape", "padding"],
    [((1, 3, 221, 234), 32), ((1, 3, 256, 256), 32), ((1, 3, 512, 512), 16), ((1, 3, 512, 512), 7)],
)
def test_pad_unpad(shape, padding):
    x = torch.randn(shape)

    x_padded, pad_params = pad_image_tensor(x, pad_size=padding)
    assert x_padded.size(2) % padding == 0
    assert x_padded.size(3) % padding == 0

    y = unpad_image_tensor(x_padded, pad_params)
    assert (x == y).all()


@pytest.mark.parametrize(["shape", "padding"], [((1, 3, 512, 512), (7, 13))])
def test_pad_unpad_nonsymmetric(shape, padding):
    x = torch.randn(shape)

    x_padded, pad_params = pad_image_tensor(x, pad_size=padding)
    assert x_padded.size(2) % padding[0] == 0
    assert x_padded.size(3) % padding[1] == 0

    y = unpad_image_tensor(x_padded, pad_params)
    assert (x == y).all()


@pytest.mark.parametrize(
    (
        "gt_bboxes",
        "pred_bboxes",
        "num_classes",
        "true_positives",
        "false_positives",
        "false_negatives",
        "confusion_matrix",
    ),
    [
        # Perfect match
        (
            ([[10, 10, 20, 20]], [0]),
            ([[10, 10, 20, 20]], [0]),
            1,
            np.array([1]),
            np.array([0]),
            np.array([0]),
            np.array([[1, 0], [0, 0]]),
        ),
        # Class mistmatch
        (
            ([[10, 10, 20, 20]], [0]),
            ([[10, 10, 20, 20]], [1]),
            2,
            np.array([0, 0]),
            np.array([0, 1]),
            np.array([1, 0]),
            np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0]]),
        ),
        # Full mismatch
        (
            ([[10, 10, 20, 20]], [0]),
            ([[30, 30, 50, 50]], [0]),
            1,
            np.array([0]),
            np.array([1]),
            np.array([1]),
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_match_bboxes(
    gt_bboxes, pred_bboxes, num_classes, true_positives, false_positives, false_negatives, confusion_matrix
):
    gt_bboxes, gt_labels = gt_bboxes
    pred_bboxes, pred_labels = pred_bboxes

    gt_bboxes = np.asarray(gt_bboxes)
    gt_labels = np.asarray(gt_labels)

    pred_bboxes = np.asarray(pred_bboxes)
    pred_labels = np.asarray(pred_labels)

    pred_scores = np.ones((len(pred_bboxes)))
    result = match_bboxes(
        pred_bboxes, pred_labels, pred_scores, gt_bboxes, gt_labels, num_classes=num_classes, iou_threshold=0.5
    )

    np.testing.assert_equal(result.true_positives, true_positives)
    np.testing.assert_equal(result.false_positives, false_positives)
    np.testing.assert_equal(result.false_negatives, false_negatives)
    np.testing.assert_equal(result.confusion_matrix, confusion_matrix)


@pytest.mark.parametrize(
    (
        "gt_bboxes",
        "pred_bboxes",
        "num_classes",
        "true_positives",
        "false_positives",
        "false_negatives",
        "confusion_matrix",
    ),
    [
        # Perfect match
        (
            ([[10, 10, 20, 20]], [0]),
            ([[10, 10, 20, 20]], [0]),
            1,
            np.array([1]),
            np.array([0]),
            np.array([0]),
            np.array([[1, 0], [0, 0]]),
        ),
        # Class mistmatch
        (
            ([[10, 10, 20, 20]], [0]),
            ([[10, 10, 20, 20]], [1]),
            2,
            np.array([0, 0]),
            np.array([0, 1]),
            np.array([1, 0]),
            np.array([[0, 1, 0], [0, 0, 0], [0, 0, 0]]),
        ),
        # Full mismatch
        (
            ([[10, 10, 20, 20]], [0]),
            ([[30, 30, 50, 50]], [0]),
            1,
            np.array([0]),
            np.array([1]),
            np.array([1]),
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_match_bboxes_hungarian(
    gt_bboxes, pred_bboxes, num_classes, true_positives, false_positives, false_negatives, confusion_matrix
):
    gt_bboxes, gt_labels = gt_bboxes
    pred_bboxes, pred_labels = pred_bboxes

    gt_bboxes = np.asarray(gt_bboxes)
    gt_labels = np.asarray(gt_labels)

    pred_bboxes = np.asarray(pred_bboxes)
    pred_labels = np.asarray(pred_labels)

    result = match_bboxes_hungarian(
        pred_bboxes, pred_labels, gt_bboxes, gt_labels, num_classes=num_classes, iou_threshold=0.5
    )

    np.testing.assert_equal(result.true_positives, true_positives)
    np.testing.assert_equal(result.false_positives, false_positives)
    np.testing.assert_equal(result.false_negatives, false_negatives)
    np.testing.assert_equal(result.confusion_matrix, confusion_matrix)


def my_collate_fn(batch, some_arg):
    return batch


class DummyDataset1(torch.utils.data.Dataset):
    def __len__(self):
        return 1

    def get_collate_fn(self):
        return partial(my_collate_fn, some_arg=1)


class DummyDataset2(torch.utils.data.Dataset):
    def __len__(self):
        return 1

    def get_collate_fn(self):
        return partial(my_collate_fn, some_arg=2)


def test_get_collate_for_dataset():
    datasets_with_different_collate = ConcatDataset([DummyDataset1(), DummyDataset2()])
    datasets_with_same_collate = ConcatDataset([DummyDataset1(), DummyDataset1()])

    with pytest.raises(ValueError):
        get_collate_for_dataset(datasets_with_different_collate, ensure_collate_fn_are_the_same=True)

    get_collate_for_dataset(datasets_with_same_collate, ensure_collate_fn_are_the_same=True)



from typing import Optional

import torch.nn.functional as F
from torch import nn, Tensor

__all__ = ["SoftBCEWithLogitsLoss"]


class SoftBCEWithLogitsLoss(nn.Module):
    """
    Drop-in replacement for nn.BCEWithLogitsLoss with few additions:
    - Support of ignore_index value
    - Support of label smoothing
    """

    __constants__ = ["weight", "pos_weight", "reduction", "ignore_index", "smooth_factor"]

    def __init__(
        self, weight=None, ignore_index: Optional[int] = -100, reduction="mean", smooth_factor=None, pos_weight=None
    ):
        super().__init__()
        self.ignore_index = ignore_index
        self.reduction = reduction
        self.smooth_factor = smooth_factor
        self.register_buffer("weight", weight)
        self.register_buffer("pos_weight", pos_weight)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if self.smooth_factor is not None:
            soft_targets = ((1 - target) * self.smooth_factor + target * (1 - self.smooth_factor)).type_as(input)
        else:
            soft_targets = target.type_as(input)

        loss = F.binary_cross_entropy_with_logits(
            input, soft_targets, self.weight, pos_weight=self.pos_weight, reduction="none"
        )

        if self.ignore_index is not None:
            not_ignored_mask: Tensor = target != self.ignore_index
            loss *= not_ignored_mask.type_as(loss)

        if self.reduction == "mean":
            loss = loss.mean()

        if self.reduction == "sum":
            loss = loss.sum()

        return loss



from collections import OrderedDict
from typing import Union, Type

import torch
from torch import nn, Tensor

from pytorch_toolbelt.modules import ACT_RELU, instantiate_activation_block, DepthwiseSeparableConv2d


class ASPPModule(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: int,
        dilation: int,
        norm_layer=nn.BatchNorm2d,
        activation=ACT_RELU,
    ):
        super(ASPPModule, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=1,
            padding=padding,
            dilation=dilation,
            bias=False,
        )
        self.abn = nn.Sequential(
            OrderedDict(
                [("norm", norm_layer(out_channels)), ("act", instantiate_activation_block(activation, inplace=True))]
            )
        )

    def forward(self, x):  # skipcq: PYL-W0221
        x = self.conv(x)
        x = self.abn(x)
        return x


class SeparableASPPModule(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        padding: int,
        dilation: int,
        norm_layer=nn.BatchNorm2d,
        activation=ACT_RELU,
    ):
        super().__init__()
        self.conv = DepthwiseSeparableConv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=1,
            padding=padding,
            dilation=dilation,
            bias=False,
        )
        self.abn = nn.Sequential(
            OrderedDict(
                [("norm", norm_layer(out_channels)), ("act", instantiate_activation_block(activation, inplace=True))]
            )
        )

    def forward(self, x):  # skipcq: PYL-W0221
        x = self.conv(x)
        x = self.abn(x)
        return x


class ASPPPooling(nn.Sequential):
    def __init__(self, in_channels: int, out_channels: int, norm_layer=nn.BatchNorm2d, activation: str = ACT_RELU):
        super().__init__()
        self.pooling = nn.AdaptiveAvgPool2d(1)
        self.conv = nn.Conv2d(in_channels, out_channels, 1, bias=False)
        self.abn = nn.Sequential(
            OrderedDict(
                [("norm", norm_layer(out_channels)), ("act", instantiate_activation_block(activation, inplace=True))]
            )
        )

    def forward(self, x: Tensor) -> Tensor:
        size = x.shape[-2:]
        x = self.pooling(x)
        x = self.conv(x)
        x = self.abn(x)
        return torch.nn.functional.interpolate(x, size=size, mode="bilinear", align_corners=False)


class ASPP(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        aspp_module: Union[Type[ASPPModule], Type[SeparableASPPModule]],
        atrous_rates=(12, 24, 36),
        dropout: float = 0.5,
        activation: str = ACT_RELU,
    ):
        super(ASPP, self).__init__()
        aspp_modules = [
            aspp_module(in_channels, out_channels, 3, padding=1, dilation=1, activation=activation),
            ASPPPooling(in_channels, out_channels),
        ] + [aspp_module(in_channels, out_channels, 3, padding=ar, dilation=ar) for ar in atrous_rates]

        self.aspp = nn.ModuleList(aspp_modules)
        self.project = nn.Sequential(
            nn.Conv2d(len(self.aspp) * out_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            instantiate_activation_block(activation, inplace=True),
            nn.Dropout2d(dropout, inplace=False),
        )

    def forward(self, x: Tensor) -> Tensor:  # skipcq: PYL-W0221
        res = []
        for aspp in self.aspp:
            res.append(aspp(x))
        res = torch.cat(res, dim=1)
        return self.project(res)



from typing import List

import torch
from torch import nn, Tensor

from pytorch_toolbelt.modules import instantiate_activation_block, GeneralizedMeanPooling2d
from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification, HasOutputFeaturesSpecification

__all__ = [
    "GlobalAveragePoolingClassificationHead",
    "GlobalMaxPoolingClassificationHead",
    "GenericPoolingClassificationHead",
    "FullyConnectedClassificationHead",
    "GlobalMaxAvgPoolingClassificationHead",
    "GlobalMaxAvgSumPoolingClassificationHead",
    "GeneralizedMeanPoolingClassificationHead",
]


class GenericPoolingClassificationHead(AbstractHead):
    def __init__(
        self,
        *,
        input_spec: FeatureMapsSpecification,
        pooling: nn.Module,
        num_classes: int,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        super().__init__(input_spec)
        self.pooling = pooling
        self.feature_map_index = feature_map_index
        self.num_classes = num_classes
        self.dropout = nn.Dropout(dropout_rate)
        self.classifier = nn.Linear(input_spec.channels[self.feature_map_index], num_classes)

    def forward(self, feature_maps: List[Tensor]) -> Tensor:
        x = feature_maps[self.feature_map_index]
        x = self.pooling(x).flatten(start_dim=1)
        x = self.dropout(x)
        x = self.classifier(x)
        return x


class GlobalMaxPoolingClassificationHead(GenericPoolingClassificationHead, HasOutputFeaturesSpecification):
    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        pooling = nn.AdaptiveMaxPool2d((1, 1))
        super().__init__(
            input_spec=input_spec,
            pooling=pooling,
            num_classes=num_classes,
            dropout_rate=dropout_rate,
            feature_map_index=feature_map_index,
        )

    def get_output_spec(self) -> FeatureMapsSpecification:
        return FeatureMapsSpecification(channels=[self.num_classes], strides=[-1])


class GlobalAveragePoolingClassificationHead(GenericPoolingClassificationHead, HasOutputFeaturesSpecification):
    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        pooling = nn.AdaptiveAvgPool2d((1, 1))

        super().__init__(
            input_spec=input_spec,
            pooling=pooling,
            num_classes=num_classes,
            dropout_rate=dropout_rate,
            feature_map_index=feature_map_index,
        )

    def get_output_spec(self) -> FeatureMapsSpecification:
        return FeatureMapsSpecification(channels=[self.num_classes], strides=[-1])


class GlobalMaxAvgPoolingClassificationHead(AbstractHead, HasOutputFeaturesSpecification):
    def __init__(
        self,
        *,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        activation: str,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        super().__init__(input_spec)
        self.num_classes = num_classes
        self.max_pooling = nn.AdaptiveMaxPool2d((1, 1))
        self.avg_pooling = nn.AdaptiveAvgPool2d((1, 1))
        self.feature_map_index = feature_map_index
        num_channels = input_spec.channels[self.feature_map_index]
        self.bottleneck = nn.Sequential(
            nn.BatchNorm1d(num_channels * 2),
            nn.Linear(num_channels * 2, num_channels),
            instantiate_activation_block(activation, inplace=True),
            nn.Dropout(dropout_rate),
            nn.BatchNorm1d(num_channels),
            nn.Linear(num_channels, num_channels),
            instantiate_activation_block(activation, inplace=True),
            nn.Dropout(dropout_rate),
        )
        self.classifier = nn.Linear(num_channels, num_classes)

    def forward(self, feature_maps: List[Tensor]) -> Tensor:
        x = feature_maps[self.feature_map_index]
        x_max = self.max_pooling(x).flatten(start_dim=1)
        x_avg = self.avg_pooling(x).flatten(start_dim=1)
        x = torch.cat([x_max, x_avg], dim=1)
        x = self.bottleneck(x)
        x = self.classifier(x)
        return x

    def get_output_spec(self) -> FeatureMapsSpecification:
        return FeatureMapsSpecification(channels=[self.num_classes], strides=[-1])


class GeneralizedMeanPoolingClassificationHead(AbstractHead, HasOutputFeaturesSpecification):
    def __init__(
        self,
        *,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        super().__init__(input_spec)
        self.num_classes = num_classes
        self.pooling = GeneralizedMeanPooling2d(l2_normalize=True, flatten=True)

        self.feature_map_index = feature_map_index
        num_channels = input_spec.channels[self.feature_map_index]
        self.dropout = nn.Dropout(dropout_rate, inplace=True)
        self.classifier = nn.Linear(num_channels, num_classes)

    def forward(self, feature_maps: List[Tensor]) -> Tensor:
        x = feature_maps[self.feature_map_index]
        x = self.pooling(x)
        x = self.dropout(x)
        x = self.classifier(x)
        return x

    def get_output_spec(self) -> FeatureMapsSpecification:
        return FeatureMapsSpecification(channels=[self.num_classes], strides=[-1])


class FullyConnectedClassificationHead(AbstractHead):
    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        dropout_rate: float = 0.0,
        feature_map_index: int = -1,
    ):
        super().__init__(input_spec)
        self.feature_map_index = feature_map_index
        self.dropout = nn.Dropout(dropout_rate)
        self.classifier = nn.LazyLinear(num_classes)

    def forward(self, feature_maps: List[Tensor]) -> Tensor:
        x = feature_maps[self.feature_map_index]
        x = torch.flatten(x, start_dim=1)
        x = self.dropout(x)
        x = self.classifier(x)
        return x


class GlobalMaxAvgSumPoolingClassificationHead(AbstractHead, HasOutputFeaturesSpecification):
    def __init__(
        self,
        *,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        feature_map_index: int = -1,
        dropout_rate: float = 0.0,
    ):
        super().__init__(input_spec)
        self.num_classes = num_classes
        self.max_pooling = nn.AdaptiveMaxPool2d((1, 1))
        self.avg_pooling = nn.AdaptiveAvgPool2d((1, 1))
        self.feature_map_index = feature_map_index
        self.num_classes = num_classes
        self.dropout = nn.Dropout(dropout_rate, inplace=True) if dropout_rate > 0 else nn.Identity()
        num_channels = input_spec.channels[self.feature_map_index]
        self.classifier = nn.Linear(num_channels, num_classes)

    def forward(self, feature_maps: List[Tensor]) -> Tensor:
        x = feature_maps[self.feature_map_index]
        x_max = self.max_pooling(x).flatten(start_dim=1)
        x_avg = self.avg_pooling(x).flatten(start_dim=1)
        x = self.dropout((x_max + x_avg) * 0.5)
        x = self.classifier(x)
        return x

    def get_output_spec(self) -> FeatureMapsSpecification:
        return FeatureMapsSpecification(channels=[self.num_classes], strides=[-1])



from collections import OrderedDict
from typing import List, Union

import torch
from torch import nn

from pytorch_toolbelt.modules import ACT_RELU, instantiate_activation_block, ChannelSpatialGate2d
from pytorch_toolbelt.modules.encoders import EncoderModule, make_n_channel_input

__all__ = [
    "XResNet18Encoder",
    "XResNet34Encoder",
    "XResNet50Encoder",
    "XResNet101Encoder",
    "XResNet152Encoder",
    "SEXResNet18Encoder",
    "SEXResNet34Encoder",
    "SEXResNet50Encoder",
    "SEXResNet101Encoder",
    "SEXResNet152Encoder",
]


def make_conv_bn_act(
    in_channels: int,
    out_channels: int,
    kernel_size: int = 3,
    stride: int = 1,
    zero_batch_norm: bool = False,
    use_activation: bool = True,
    activation: str = ACT_RELU,
) -> torch.nn.Sequential:
    """
    Create a nn.Conv2d block followed by nn.BatchNorm2d and (optional) activation block.
    """
    batch_norm = nn.BatchNorm2d(out_channels)
    # initializer batch normalization to 0 if its the final conv layer
    nn.init.constant_(batch_norm.weight, 0.0 if zero_batch_norm else 1.0)
    layers = [
        (
            "conv",
            nn.Conv2d(in_channels, out_channels, kernel_size, stride=stride, padding=kernel_size // 2, bias=False),
        ),
        ("bn", batch_norm),
    ]
    if use_activation:
        activation_block = instantiate_activation_block(activation, inplace=True)
        layers.append((activation, activation_block))

    return nn.Sequential(OrderedDict(layers))


class StemBlock(nn.Module):
    def __init__(self, input_channels: int, output_channels: int, activation: str = ACT_RELU):
        super().__init__()
        self.conv_bn_relu_1 = make_conv_bn_act(input_channels, 8, stride=2, activation=activation)
        self.conv_bn_relu_2 = make_conv_bn_act(8, 64, activation=activation)
        self.conv_bn_relu_3 = make_conv_bn_act(64, output_channels, activation=activation)

    def forward(self, x):
        x = self.conv_bn_relu_1(x)
        x = self.conv_bn_relu_2(x)
        x = self.conv_bn_relu_3(x)
        return x


class XResNetBlock(nn.Module):
    """Creates the standard `XResNet` block."""

    def __init__(self, expansion: int, n_inputs: int, n_hidden: int, stride: int = 1, activation: str = ACT_RELU):
        super().__init__()

        n_inputs = n_inputs * expansion
        n_filters = n_hidden * expansion

        # convolution path
        if expansion == 1:
            layers = [
                make_conv_bn_act(n_inputs, n_hidden, 3, stride=stride, activation=activation),
                make_conv_bn_act(n_hidden, n_filters, 3, zero_batch_norm=True, use_activation=False),
            ]
        else:
            layers = [
                make_conv_bn_act(n_inputs, n_hidden, 1, activation=activation),
                make_conv_bn_act(n_hidden, n_hidden, 3, stride=stride, activation=activation),
                make_conv_bn_act(n_hidden, n_filters, 1, zero_batch_norm=True, use_activation=False),
            ]

        self.convs = nn.Sequential(*layers)
        self.activation = instantiate_activation_block(activation, inplace=True)

        # identity path
        if n_inputs == n_filters:
            self.id_conv = nn.Identity()
        else:
            self.id_conv = make_conv_bn_act(n_inputs, n_filters, kernel_size=1, use_activation=False)
        if stride == 1:
            self.pool = nn.Identity()
        else:
            self.pool = nn.AvgPool2d(2, ceil_mode=True)

    def forward(self, x):
        return self.activation(self.convs(x) + self.id_conv(self.pool(x)))


class SEXResNetBlock(nn.Module):
    """Creates the Squeeze&Excitation + XResNet block."""

    def __init__(self, expansion: int, n_inputs: int, n_hidden: int, stride: int = 1, activation: str = ACT_RELU):
        super().__init__()

        n_inputs = n_inputs * expansion
        n_filters = n_hidden * expansion

        # convolution path
        if expansion == 1:
            layers = [
                make_conv_bn_act(n_inputs, n_hidden, 3, stride=stride, activation=activation),
                make_conv_bn_act(n_hidden, n_filters, 3, zero_batch_norm=True, use_activation=False),
            ]
        else:
            layers = [
                make_conv_bn_act(n_inputs, n_hidden, 1, activation=activation),
                make_conv_bn_act(n_hidden, n_hidden, 3, stride=stride, activation=activation),
                make_conv_bn_act(n_hidden, n_filters, 1, zero_batch_norm=True, use_activation=False),
            ]

        self.convs = nn.Sequential(*layers)
        self.activation = instantiate_activation_block(activation, inplace=True)
        self.se = ChannelSpatialGate2d(n_filters, reduction=4)

        # identity path
        if n_inputs == n_filters:
            self.id_conv = nn.Identity()
        else:
            self.id_conv = make_conv_bn_act(n_inputs, n_filters, kernel_size=1, use_activation=False)
        if stride == 1:
            self.pool = nn.Identity()
        else:
            self.pool = nn.AvgPool2d(2, ceil_mode=True)

    def forward(self, x):
        return self.activation(self.convs(x) + self.id_conv(self.pool(x)))


class XResNet(EncoderModule):
    def __init__(
        self,
        expansion: int,
        blocks: List[int],
        input_channels: int = 3,
        activation=ACT_RELU,
        layers=None,
        first_pool: Union[nn.MaxPool2d, nn.AvgPool2d] = nn.MaxPool2d,
        pretrained=None,
        block: Union[XResNetBlock, SEXResNetBlock] = XResNetBlock,
    ):
        assert len(blocks) == 4
        if layers is None:
            layers = [1, 2, 3, 4]

        n_filters = [64 // expansion, 64, 128, 256, 512]
        channels = [64, 64 * expansion, 128 * expansion, 256 * expansion, 512 * expansion]

        super().__init__(channels, [2, 4, 8, 16, 32], layers)

        res_layers = [
            self._make_layer(
                block,
                expansion,
                n_filters[i],
                n_filters[i + 1],
                n_blocks=l,
                stride=1 if i == 0 else 2,
                activation=activation,
            )
            for i, l in enumerate(blocks)
        ]

        self.stem = StemBlock(input_channels, 64, activation=activation)
        self.layer1 = nn.Sequential(
            OrderedDict([("pool", first_pool(kernel_size=3, stride=2, padding=1)), ("block", res_layers[0])])
        )

        self.layer2 = res_layers[1]
        self.layer3 = res_layers[2]
        self.layer4 = res_layers[3]

    @property
    def encoder_layers(self) -> List[nn.Module]:
        return [self.stem, self.layer1, self.layer2, self.layer3, self.layer4]

    @staticmethod
    def _make_layer(block, expansion, n_inputs: int, n_filters: int, n_blocks: int, stride: int, activation: str):
        return nn.Sequential(
            *[
                block(
                    expansion,
                    n_inputs if i == 0 else n_filters,
                    n_filters,
                    stride if i == 0 else 1,
                    activation=activation,
                )
                for i in range(n_blocks)
            ]
        )

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.stem.conv_bn_relu_1.conv = make_n_channel_input(self.stem.conv_bn_relu_1.conv, input_channels, mode)


def XResNet18Encoder(**kwargs):
    return XResNet(1, [2, 2, 2, 2], **kwargs)


def XResNet34Encoder(**kwargs):
    return XResNet(1, [3, 4, 6, 3], **kwargs)


def XResNet50Encoder(**kwargs):
    return XResNet(4, [3, 4, 6, 3], **kwargs)


def XResNet101Encoder(**kwargs):
    return XResNet(4, [3, 4, 23, 3], **kwargs)


def XResNet152Encoder(**kwargs):
    return XResNet(4, [3, 8, 36, 3], **kwargs)


# SE-XResNet


def SEXResNet18Encoder(**kwargs):
    return XResNet(1, [2, 2, 2, 2], block=SEXResNetBlock, **kwargs)


def SEXResNet34Encoder(**kwargs):
    return XResNet(1, [3, 4, 6, 3], block=SEXResNetBlock, **kwargs)


def SEXResNet50Encoder(**kwargs):
    return XResNet(4, [3, 4, 6, 3], block=SEXResNetBlock, **kwargs)


def SEXResNet101Encoder(**kwargs):
    return XResNet(4, [3, 4, 23, 3], block=SEXResNetBlock, **kwargs)


def SEXResNet152Encoder(**kwargs):
    return XResNet(4, [3, 8, 36, 3], block=SEXResNetBlock, **kwargs)



from pprint import pprint

import pytest
import torch

import pytorch_toolbelt.modules.decoders as D
from pytorch_toolbelt.modules.interfaces import FeatureMapsSpecification
from pytorch_toolbelt.modules.upsample import UpsampleLayerType
from pytorch_toolbelt.utils.torch_utils import count_parameters, describe_outputs


@torch.no_grad()
@pytest.mark.parametrize(
    ("decoder_cls", "decoder_params"),
    [
        (D.FPNDecoder, {"out_channels": 128, "upsample_block": UpsampleLayerType.BILINEAR}),
        (D.FPNDecoder, {"out_channels": 128, "upsample_block": UpsampleLayerType.NEAREST}),
        (D.FPNDecoder, {"out_channels": 128, "upsample_block": UpsampleLayerType.DECONVOLUTION}),
        (D.BiFPNDecoder, {"out_channels": 128, "num_layers": 3}),
        (D.DeeplabV3PlusDecoder, {"out_channels": 128, "aspp_channels": 256}),
        (D.DeeplabV3Decoder, {"out_channels": 128, "aspp_channels": 256}),
        (D.CANDecoder, {"out_channels": 128}),
        #
        (D.UNetDecoder, {"out_channels": [128, 256, 384, 512], "upsample_block": UpsampleLayerType.DECONVOLUTION}),
        (D.UNetDecoder, {"out_channels": [128, 256, 384, 512], "upsample_block": UpsampleLayerType.NEAREST}),
        (D.UNetDecoder, {"out_channels": [128, 256, 384, 512], "upsample_block": UpsampleLayerType.BILINEAR}),
        (D.UNetDecoder, {"out_channels": [128, 256, 384, 512], "upsample_block": UpsampleLayerType.PIXEL_SHUFFLE}),
        (D.UNetDecoder, {"out_channels": [128, 256, 384, 512], "upsample_block": UpsampleLayerType.RESIDUAL_DECONV}),
    ],
)
def test_decoders(decoder_cls, decoder_params):
    input_spec = FeatureMapsSpecification(channels=(64, 128, 256, 512, 1024), strides=(4, 8, 16, 32, 64))
    input = input_spec.get_dummy_input()

    decoder = decoder_cls(input_spec, **decoder_params).eval()
    output = decoder(input)

    print()
    print(decoder.__class__.__name__)
    print(count_parameters(decoder, human_friendly=True))
    pprint(describe_outputs(output))

    torch.jit.trace(decoder, (input,))



from typing import Union

import cv2

__all__ = [
    "INPUT_IMAGE_ID_KEY",
    "INPUT_IMAGE_KEY",
    "INPUT_INDEX_KEY",
    "OUTPUT_EMBEDDINGS_KEY",
    "OUTPUT_LOGITS_KEY",
    "OUTPUT_MASK_KEY",
    "OUTPUT_MASK_KEY_STRIDE_16",
    "OUTPUT_MASK_KEY_STRIDE_2",
    "OUTPUT_MASK_KEY_STRIDE_32",
    "OUTPUT_MASK_KEY_STRIDE_4",
    "OUTPUT_MASK_KEY_STRIDE_64",
    "OUTPUT_MASK_KEY_STRIDE_8",
    "TARGET_CLASS_KEY",
    "TARGET_LABELS_KEY",
    "TARGET_MASK_KEY",
    "TARGET_MASK_KEY_STRIDE_16",
    "TARGET_MASK_KEY_STRIDE_2",
    "TARGET_MASK_KEY_STRIDE_32",
    "TARGET_MASK_KEY_STRIDE_4",
    "TARGET_MASK_KEY_STRIDE_64",
    "TARGET_MASK_KEY_STRIDE_8",
    "TARGET_MASK_WEIGHT_KEY",
    "name_for_stride",
    "read_image_rgb",
]


def name_for_stride(name, stride: Union[int, None]) -> str:
    if stride is None:
        return name
    return f"{name}_STRIDE_{stride}"


INPUT_INDEX_KEY = "INPUT_INDEX_KEY"
INPUT_IMAGE_KEY = "INPUT_IMAGE_KEY"
INPUT_IMAGE_ID_KEY = "INPUT_IMAGE_ID_KEY"

TARGET_MASK_WEIGHT_KEY = "TARGET_MASK_WEIGHT_KEY"
TARGET_CLASS_KEY = "TARGET_CLASS_KEY"
TARGET_LABELS_KEY = "TARGET_LABELS_KEY"

TARGET_MASK_KEY = "TARGET_MASK_KEY"

TARGET_MASK_KEY_STRIDE_2 = name_for_stride(TARGET_MASK_KEY, 2)
TARGET_MASK_KEY_STRIDE_4 = name_for_stride(TARGET_MASK_KEY, 4)
TARGET_MASK_KEY_STRIDE_8 = name_for_stride(TARGET_MASK_KEY, 8)
TARGET_MASK_KEY_STRIDE_16 = name_for_stride(TARGET_MASK_KEY, 16)
TARGET_MASK_KEY_STRIDE_32 = name_for_stride(TARGET_MASK_KEY, 32)
TARGET_MASK_KEY_STRIDE_64 = name_for_stride(TARGET_MASK_KEY, 64)

OUTPUT_MASK_KEY = "OUTPUT_MASK_KEY"
OUTPUT_MASK_KEY_STRIDE_2 = name_for_stride(OUTPUT_MASK_KEY, 2)
OUTPUT_MASK_KEY_STRIDE_4 = name_for_stride(OUTPUT_MASK_KEY, 4)
OUTPUT_MASK_KEY_STRIDE_8 = name_for_stride(OUTPUT_MASK_KEY, 8)
OUTPUT_MASK_KEY_STRIDE_16 = name_for_stride(OUTPUT_MASK_KEY, 16)
OUTPUT_MASK_KEY_STRIDE_32 = name_for_stride(OUTPUT_MASK_KEY, 32)
OUTPUT_MASK_KEY_STRIDE_64 = name_for_stride(OUTPUT_MASK_KEY, 64)

OUTPUT_LOGITS_KEY = "OUTPUT_LOGITS_KEY"
OUTPUT_EMBEDDINGS_KEY = "OUTPUT_EMBEDDINGS_KEY"


def read_image_rgb(fname: str):
    image = cv2.imread(fname)[..., ::-1]
    if image is None:
        raise IOError("Cannot read " + fname)
    return image



import pytest
import torch
from pytorch_toolbelt.modules import HFF, ResidualDeconvolutionUpsample2d, GlobalKMaxPool2d

skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="Cuda is not available")


def test_hff_dynamic_size():
    feature_maps = [
        torch.randn((4, 3, 512, 512)),
        torch.randn((4, 3, 256, 256)),
        torch.randn((4, 3, 128, 128)),
        torch.randn((4, 3, 64, 64)),
    ]

    hff = HFF(upsample_scale=2)
    output = hff(feature_maps)
    assert output.size(2) == 512
    assert output.size(3) == 512


def test_hff_static_size():
    feature_maps = [
        torch.randn((4, 3, 512, 512)),
        torch.randn((4, 3, 384, 384)),
        torch.randn((4, 3, 256, 256)),
        torch.randn((4, 3, 128, 128)),
        torch.randn((4, 3, 32, 32)),
    ]

    hff = HFF(sizes=[(512, 512), (384, 384), (256, 256), (128, 128), (32, 32)])
    output = hff(feature_maps)
    assert output.size(2) == 512
    assert output.size(3) == 512


# def test_upsample():
#     block = DepthToSpaceUpsample2d(1)
#     original = np.expand_dims(cv2.imread("lena.png", cv2.IMREAD_GRAYSCALE), -1)
#     input = tensor_from_rgb_image(original / 255.0).unsqueeze(0).float()
#     output = block(input)
#
#     output_rgb = rgb_image_from_tensor(output.squeeze(0), mean=0, std=1, max_pixel_value=1, dtype=np.float32)
#
#     cv2.imshow("Original", original)
#     cv2.imshow("Upsampled (cv2)", cv2.resize(original, None, fx=2, fy=2, interpolation=cv2.INTER_LINEAR))
#     cv2.imshow("Upsampled", cv2.normalize(output_rgb, None, 0, 255, cv2.NORM_MINMAX, dtype=cv2.CV_8U))
#     cv2.waitKey(-1)


def test_residualdeconvolutionupsampleblock():
    x = torch.randn((4, 16, 32, 32))
    block = ResidualDeconvolutionUpsample2d(16)
    output = block(x, output_size=None)
    print(x.size(), x.mean(), x.std())
    print(output.size(), output.mean(), x.std())


def test_kmax_pool():
    x = torch.randn((8, 512, 16, 16))
    module1 = GlobalKMaxPool2d(k=4, flatten=True)
    module2 = GlobalKMaxPool2d(k=4, flatten=False)

    y1 = module1(x)
    y2 = module2(x)

    assert y1.size() == (8, 512)
    assert y2.size() == (8, 512, 1, 1)



import torch
from torch import nn, Tensor
from typing import Optional

__all__ = ["soft_micro_f1", "BinarySoftF1Loss", "SoftF1Loss"]


def soft_micro_f1(preds: Tensor, targets: Tensor, eps=1e-6) -> Tensor:
    """Compute the macro soft F1-score as a cost.
    Average (1 - soft-F1) across all labels.
    Use probability values instead of binary predictions.

    Args:
        targets (Tensor): targets array of shape (Num Samples, Num Classes)
        preds (Tensor): probability matrix of shape (Num Samples, Num Classes)

    Returns:
        cost (scalar Tensor): value of the cost function for the batch

    References:
        https://towardsdatascience.com/the-unknown-benefits-of-using-a-soft-f1-loss-in-classification-systems-753902c0105d
    """
    tp = torch.sum(preds * targets, dim=0)
    fp = torch.sum(preds * (1 - targets), dim=0)
    fn = torch.sum((1 - preds) * targets, dim=0)
    soft_f1 = 2 * tp / (2 * tp + fn + fp + eps)
    loss = 1 - soft_f1  # reduce 1 - soft-f1 in order to increase soft-f1
    return loss.mean()


# TODO: Test
# def macro_double_soft_f1(y, y_hat):
#     """Compute the macro soft F1-score as a cost (average 1 - soft-F1 across all labels).
#     Use probability values instead of binary predictions.
#     This version uses the computation of soft-F1 for both positive and negative class for each label.
#
#     Args:
#         y (int32 Tensor): targets array of shape (BATCH_SIZE, N_LABELS)
#         y_hat (float32 Tensor): probability matrix from forward propagation of shape (BATCH_SIZE, N_LABELS)
#
#     Returns:
#         cost (scalar Tensor): value of the cost function for the batch
#     """
#     tp = tf.reduce_sum(y_hat * y, axis=0)
#     fp = tf.reduce_sum(y_hat * (1 - y), axis=0)
#     fn = tf.reduce_sum((1 - y_hat) * y, axis=0)
#     tn = tf.reduce_sum((1 - y_hat) * (1 - y), axis=0)
#     soft_f1_class1 = 2 * tp / (2 * tp + fn + fp + 1e-16)
#     soft_f1_class0 = 2 * tn / (2 * tn + fn + fp + 1e-16)
#     cost_class1 = 1 - soft_f1_class1  # reduce 1 - soft-f1_class1 in order to increase soft-f1 on class 1
#     cost_class0 = 1 - soft_f1_class0  # reduce 1 - soft-f1_class0 in order to increase soft-f1 on class 0
#     cost = 0.5 * (cost_class1 + cost_class0)  # take into account both class 1 and class 0
#     macro_cost = tf.reduce_mean(cost)  # average on all labels
#     return macro_cost


class BinarySoftF1Loss(nn.Module):
    def __init__(self, ignore_index: Optional[int] = None, eps=1e-6):
        super().__init__()
        self.ignore_index = ignore_index
        self.eps = eps

    def forward(self, preds: Tensor, targets: Tensor) -> Tensor:
        targets = targets.view(-1)
        preds = preds.view(-1)

        if self.ignore_index is not None:
            # Filter predictions with ignore label from loss computation
            not_ignored = targets != self.ignore_index
            preds = preds[not_ignored]
            targets = targets[not_ignored]

            if targets.numel() == 0:
                return torch.tensor(0, dtype=preds.dtype, device=preds.device)

        preds = preds.sigmoid().clamp(self.eps, 1 - self.eps)
        return soft_micro_f1(preds.view(-1, 1), targets.view(-1, 1))


class SoftF1Loss(nn.Module):
    def __init__(self, ignore_index: Optional[int] = None, eps=1e-6):
        super().__init__()
        self.ignore_index = ignore_index
        self.eps = eps

    def forward(self, preds: Tensor, targets: Tensor) -> Tensor:
        preds = preds.softmax(dim=1).clamp(self.eps, 1 - self.eps)
        targets = torch.nn.functional.one_hot(targets, preds.size(1))

        if self.ignore_index is not None:
            # Filter predictions with ignore label from loss computation
            not_ignored = targets != self.ignore_index
            preds = preds[not_ignored]
            targets = targets[not_ignored]

            if targets.numel() == 0:
                return torch.tensor(0, dtype=preds.dtype, device=preds.device)

        return soft_micro_f1(preds, targets)



import gc
import inspect
import logging
import os
import pickle
import typing
import functools
from typing import Any, Dict, List, Optional
import numpy as np
import torch
from torch import Tensor

import torch.distributed as dist

from pytorch_toolbelt.utils.bucket_assignment import (
    naive_bucket_assignment,
    compute_bucket_imbalance_score,
    random_bucket_assignment,
    filler_bucket_assignment,
)

__all__ = [
    "distributed_guard",
    "DistributedGuard",
    "all_gather",
    "broadcast_from_master",
    "get_rank",
    "get_world_size",
    "is_dist_avail_and_initialized",
    "is_main_process",
    "master_print",
    "reduce_dict_sum",
    "split_across_nodes",
    "master_node_only",
]

logger = logging.getLogger("pytorch_toolbelt.utils.distributed")


class DistributedGuard:
    def __init__(
        self,
        local_rank: int = os.environ.get("LOCAL_RANK", 0),
        world_size: int = os.environ.get("WORLD_SIZE", 1),
        visible_devices: List = os.environ.get("CUDA_VISIBLE_DEVICES", list(range(torch.cuda.device_count()))),
    ):
        self.local_rank = int(local_rank)
        self.world_size = int(world_size)
        self.visible_devices = visible_devices
        self.dist_is_available = torch.distributed.is_available()
        self.dist_is_initialized = torch.distributed.is_initialized()
        self.device = torch.device(f"cuda:{self.local_rank}")

        logger.info(
            f"Creating DistributedGuard with devices {self.visible_devices} {self.local_rank}/{self.world_size}"
        )

    def __enter__(self):
        torch.cuda.set_device(self.device)

        if self.dist_is_available and self.world_size > 1:
            if not self.dist_is_initialized:
                logger.info(f"Setting CUDA device {self.device} for rank {self.local_rank}/{self.world_size}")
                torch.distributed.init_process_group(backend="nccl", world_size=self.world_size, rank=self.local_rank)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if self.dist_is_available and self.dist_is_initialized:
                torch.distributed.barrier()
                # torch.distributed.destroy_process_group()
        except Exception as e:
            logger.exception(e)
        finally:
            torch.cuda.empty_cache()
            gc.collect()


def distributed_guard(func):
    def inner1(*args, **kwargs):
        with DistributedGuard():
            return func(*args, **kwargs)

    return inner1


def is_dist_avail_and_initialized() -> bool:
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size() -> int:
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank() -> int:
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process() -> bool:
    return get_rank() == 0


def broadcast_from_master(data: Any) -> Any:
    """
    Broadcast data from master node to all other nodes. This may be required when you
    want to compute something only on master node (e.g computational-heavy metric) and
    don't want to vaste CPU of other nodes doing same work simultaneously.

    >>> if is_main_process():
    >>>    result = some_code_to_run(...)
    >>> else:
    >>>    result = None
    >>> # 'result' propagated to all nodes from master
    >>> result = broadcast_from_master(result)

    Args:
        data: Data to be broadcasted from master node (rank 0)

    Returns:
        Data from rank 0 node
    """
    world_size = get_world_size()
    if world_size == 1:
        return data

    local_rank = get_rank()
    storage: Tensor

    if local_rank == 0:
        buffer = pickle.dumps(data)
        storage = torch.UntypedStorage.from_buffer(buffer, dtype=torch.uint8)
        payload = torch.ByteTensor(storage).to("cuda")
        local_size = payload.numel()
    else:
        local_size = 0

    # Propagate target tensor size to all nodes
    local_size = max(all_gather(local_size))
    if local_rank != 0:
        payload = torch.empty((local_size,), dtype=torch.uint8, device="cuda")

    dist.broadcast(payload, 0)
    buffer = payload.cpu().numpy().tobytes()
    return pickle.loads(buffer)


def all_gather(data: Any) -> List[Any]:
    """
    Run all_gather on arbitrary picklable data (not necessarily tensors)
    Args:
        data: any picklable object
    Returns:
        list[data]: list of data gathered from each rank
    """
    world_size = get_world_size()
    if world_size == 1:
        return [data]

    # serialized to a Tensor
    buffer = pickle.dumps(data)
    storage = torch.UntypedStorage.from_buffer(buffer, dtype=torch.uint8)
    tensor = torch.ByteTensor(storage).to("cuda")

    # obtain Tensor size of each rank
    local_size = torch.tensor([tensor.numel()], device="cuda")
    size_list = [torch.tensor([0], device="cuda") for _ in range(world_size)]
    dist.all_gather(size_list, local_size)
    size_list = [int(size.item()) for size in size_list]
    max_size = max(size_list)

    # receiving Tensor from all ranks
    # we pad the tensor because torch all_gather does not support
    # gathering tensors of different shapes
    tensor_list = []
    for _ in size_list:
        tensor_list.append(torch.empty((max_size,), dtype=torch.uint8, device="cuda"))
    if local_size != max_size:
        padding = torch.empty(size=(max_size - local_size,), dtype=torch.uint8, device="cuda")
        tensor = torch.cat((tensor, padding), dim=0)
    dist.all_gather(tensor_list, tensor)

    data_list = []
    for size, tensor in zip(size_list, tensor_list):
        buffer = tensor.cpu().numpy().tobytes()[:size]
        data_list.append(pickle.loads(buffer))

    return data_list


def reduce_dict_sum(input_dict: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Reduce the values in the dictionary from all processes so that all processes
    have the sum of the all values. Returns a dict with the same fields as
    input_dict, after reduction.

    Note: This function can work on defaultdict(list) and will effectively
    concatenate multiple dictionaries into single one, thanks to that summation
    of lists works as a concatenation.
    Args:
        input_dict (dict): all the values will be reduced
    """
    world_size = get_world_size()
    if world_size < 2:
        return input_dict

    with torch.no_grad():
        reduced_dict = {}
        for x in all_gather(input_dict):
            for key, value in x.items():
                if key in reduced_dict:
                    reduced_dict[key] += value
                else:
                    reduced_dict[key] = value
    return reduced_dict


def master_print(*args, **kwargs) -> None:
    """
    Drop-in replacement for built-in `print` function that prints only on master node

    Args:
        *args:
        **kwargs:

    Returns:
        None
    """
    if is_main_process():
        print(*args, **kwargs)


def split_across_nodes(
    collection: List,
    world_size: Optional[int] = None,
    local_rank: Optional[int] = None,
    cost: Optional[List] = None,
    method: str = "optimal",
) -> List:
    """
    Split input collection such that each node receives 1/N of the total collection elements to process, where
    N is the number of nodes.

    Example:

    >>> local_values = split_across_nodes([0,1,2,3,4,5,6,7,8,9])
    >>> print(local_values, get_rank())
    >>> # [0,1,2], 0
    >>> # [3,4,5], 1
    >>> # [6,7,8], 2
    >>> # [9], 3

    Args:
        collection: Initial collection of size N to split into K nodes
        world_size: World size (Number of nodes K)
        local_rank: Current node
        cost: A vector of size N that represents the cost of processing associated with each item.
              If present, it will affect the order of elements each node will receive to even the total cost each node
              will get.
        method: Bucket assignment method used to assign each sample of associated cost to specific node to minimze std of total cost per node.

                naive - Sort elements by cost then assing them in repeating patterm: [0, 1, 2, 3, 0, 1, 2, 3, 4, ...].
                Each node gets exactly the same number of samples, however cost per node may vary greatly.

                optimal - Iteratively assigns elements starting with the most costly ones to the least used bucket.
                This gives much better cost balance per node, but the number of items each node gets it not guaratneed to be equal at all.
    Returns:

    """
    if world_size is None:
        world_size = get_world_size()
    if local_rank is None:
        local_rank = get_rank()

    if world_size > 1:
        if cost is not None:
            if len(cost) != len(collection):
                raise RuntimeError()

            method_fn = {
                "optimal": filler_bucket_assignment,
                "naive": naive_bucket_assignment,
            }[method]
            assigned_indexes = method_fn(cost, world_size)

            rank_local_indexes = assigned_indexes == local_rank

            logger.debug(
                f"Node {local_rank} get {np.count_nonzero(rank_local_indexes)} items with total cost {sum(cost[rank_local_indexes])}"
            )

            if isinstance(collection, np.ndarray):
                rank_specific_subset = collection[rank_local_indexes]
            else:
                rank_specific_subset = [
                    collection[index] for index, should_pick in enumerate(rank_local_indexes) if should_pick
                ]

        else:
            indexes = np.linspace(0, len(collection), int(world_size + 1), dtype=int)
            rank_local_indexes = slice(indexes[local_rank], indexes[local_rank + 1])
            rank_specific_subset = collection[rank_local_indexes]

            logger.debug(
                f"split_across_nodes returning slice {rank_local_indexes} from collection of size {len(collection)} for rank {local_rank}"
            )
        return rank_specific_subset
    else:
        return collection


def master_node_only(func):
    """
    A decorator for making sure a function runs only in main process.
    If not in DDP mode (local_rank = -1), the function will run.
    If in DDP mode, the function will run only in the main process (local_rank = 0)
    This works only for functions with no return value
    """

    return_type = inspect.signature(func).return_annotation
    function_has_return_value = return_type is not None and return_type != inspect._empty
    if function_has_return_value:
        raise RuntimeError(
            f"Function {func} decorated with @master_node_only must not return any value. "
            f"Function signature: {inspect.signature(func)}"
        )

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if is_main_process():
            return func(*args, **kwargs)
        else:
            return None

    return wrapper



import random
import typing
from typing import Any, Optional

from torch.utils.data import Dataset
import numpy as np

__all__ = ["RandomSubsetDataset", "RandomSubsetWithMaskDataset"]

from torch.utils.data.dataloader import default_collate


class RandomSubsetDataset(Dataset):
    """
    Wrapper to get desired number of samples from underlying dataset
    """

    def __init__(self, dataset, num_samples: int, weights: Optional[np.ndarray] = None):
        if weights is not None:
            if len(dataset) != len(weights):
                raise ValueError(
                    "Length of weights must be equal to length of dataset. Got {} and {}".format(
                        len(weights), len(dataset)
                    )
                )
        self.dataset = dataset
        self.num_samples = num_samples
        self.weights = np.cumsum(weights) if weights is not None else None

    def __len__(self) -> int:
        return self.num_samples

    def __getitem__(self, _) -> Any:
        if self.weights is not None:
            population = range(len(self.dataset))
            index = random.choices(population, cum_weights=self.weights, k=1)[0]
        else:
            index = random.randrange(len(self.dataset))
        return self.dataset[index]

    def get_collate_fn(self):
        get_collate_fn = getattr(self.dataset, "get_collate_fn", None)
        if callable(get_collate_fn):
            return get_collate_fn()
        return default_collate


class RandomSubsetWithMaskDataset(Dataset):
    """
    Wrapper to get desired number of samples from underlying dataset only considering
    samples P for which mask[P] equals True
    """

    def __init__(self, dataset: Dataset, mask: np.ndarray, num_samples: int):
        if (
            not isinstance(mask, np.ndarray)
            or mask.dtype != np.bool
            or len(mask.shape) != 1
            or len(mask) != len(dataset)
        ):
            raise ValueError("Mask must be boolean 1-D numpy array")

        if not mask.any():
            raise ValueError("Mask must have at least one positive value")

        self.dataset = dataset
        self.mask = mask
        self.num_samples = num_samples
        self.indexes = np.flatnonzero(self.mask)

    def __len__(self) -> int:
        return self.num_samples

    def __getitem__(self, _) -> Any:
        index = random.choice(self.indexes)
        return self.dataset[index]

    def get_collate_fn(self) -> typing.Callable:
        get_collate_fn = getattr(self.dataset, "get_collate_fn", None)
        if callable(get_collate_fn):
            return get_collate_fn()
        return default_collate



from typing import List, Optional

import torch
import torch.nn as nn
from torch import Tensor
from pytorch_toolbelt.datasets import name_for_stride
from pytorch_toolbelt.modules import instantiate_activation_block, ACT_GELU
from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification


__all__ = ["SegFormerHead"]


class SegFormerHead(AbstractHead):
    """
    SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        embedding_dim: int,
        with_supervision: bool,
        output_name: Optional[str],
        dropout_rate: float = 0.0,
        activation: str = ACT_GELU,
    ):
        super().__init__(input_spec)

        self.num_classes = num_classes
        self.output_name = output_name
        self.with_supervision = with_supervision

        c1_in_channels, c2_in_channels, c3_in_channels, c4_in_channels = input_spec.channels

        self.linear_c4 = nn.Conv2d(c4_in_channels, embedding_dim, kernel_size=1)
        self.linear_c3 = nn.Conv2d(c3_in_channels, embedding_dim, kernel_size=1)
        self.linear_c2 = nn.Conv2d(c2_in_channels, embedding_dim, kernel_size=1)
        self.linear_c1 = nn.Conv2d(c1_in_channels, embedding_dim, kernel_size=1)

        self.linear_fuse = nn.Sequential(
            nn.Conv2d(embedding_dim * 4, embedding_dim, kernel_size=1, bias=False),
            nn.BatchNorm2d(embedding_dim),
            instantiate_activation_block(activation, inplace=True),
        )

        self.dropout = nn.Dropout2d(dropout_rate, inplace=False)
        self.final = nn.Conv2d(embedding_dim, num_classes, kernel_size=1)

        if self.with_supervision:
            self.supervision_c4 = nn.Conv2d(embedding_dim, num_classes, kernel_size=1)
            self.supervision_c3 = nn.Conv2d(embedding_dim, num_classes, kernel_size=1)
            self.supervision_c2 = nn.Conv2d(embedding_dim, num_classes, kernel_size=1)
            self.supervision_c1 = nn.Conv2d(embedding_dim, num_classes, kernel_size=1)
        else:
            self.supervision_c4 = None
            self.supervision_c3 = None
            self.supervision_c2 = None
            self.supervision_c1 = None

    def forward(self, feature_maps: List[Tensor], output_size):
        c1, c2, c3, c4 = feature_maps

        c4 = self.linear_c4(c4)
        c3 = self.linear_c3(c3)
        c2 = self.linear_c2(c2)
        c1 = self.linear_c1(c1)

        c = self.linear_fuse(
            torch.cat(
                [
                    torch.nn.functional.interpolate(c4, size=c1.size()[2:], mode="bilinear", align_corners=False),
                    torch.nn.functional.interpolate(c3, size=c1.size()[2:], mode="bilinear", align_corners=False),
                    torch.nn.functional.interpolate(c2, size=c1.size()[2:], mode="bilinear", align_corners=False),
                    c1,
                ],
                dim=1,
            )
        )

        x = self.dropout(c)
        x = self.final(x)
        x = torch.nn.functional.interpolate(x, size=output_size, mode="bilinear", align_corners=False)

        if self.output_name is not None:
            outputs = {self.output_name: x}
        else:
            outputs = x

        if self.with_supervision:
            s4, s3, s2, s1 = (
                self.supervision_c4(c4),
                self.supervision_c3(c3),
                self.supervision_c2(c2),
                self.supervision_c1(c1),
            )

            if self.output_name is not None:
                outputs[name_for_stride(self.output_name, 32)] = s4
                outputs[name_for_stride(self.output_name, 16)] = s3
                outputs[name_for_stride(self.output_name, 8)] = s2
                outputs[name_for_stride(self.output_name, 4)] = s1
            else:
                outputs = outputs + (s1, s2, s3, s4)

        return outputs

    def apply_to_final_layer(self, func):
        func(self.final)
        if self.with_supervision:
            func(self.supervision_c4)
            func(self.supervision_c3)
            func(self.supervision_c2)
            func(self.supervision_c1)



import numpy as np
import torch
from pytorch_toolbelt.inference.tiles import ImageSlicer, TileMerger
from pytorch_toolbelt.utils.torch_utils import tensor_from_rgb_image, rgb_image_from_tensor, to_numpy
from torch import nn
from torch.utils.data import DataLoader
import pytest


skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="Cuda is not available")


def test_tiles_split_merge():
    image = np.random.random((500, 500, 3)).astype(np.uint8)
    tiler = ImageSlicer(image.shape, tile_size=51, tile_step=26, weight="mean")
    tiles = tiler.split(image)
    merged = tiler.merge(tiles, dtype=np.uint8)
    np.testing.assert_equal(merged, image)


def test_tiles_split_merge_non_dividable():
    image = np.random.random((563, 512, 3)).astype(np.uint8)
    tiler = ImageSlicer(image.shape, tile_size=(128, 128), tile_step=(128, 128), weight="mean")
    tiles = tiler.split(image)
    merged = tiler.merge(tiles, dtype=np.uint8)
    np.testing.assert_equal(merged, image)


@skip_if_no_cuda
def test_tiles_split_merge_non_dividable_cuda():
    image = np.random.random((5632, 5120, 3)).astype(np.uint8)
    tiler = ImageSlicer(image.shape, tile_size=(1280, 1280), tile_step=(1280, 1280), weight="mean")
    tiles = tiler.split(image)

    merger = TileMerger(tiler.target_shape, channels=image.shape[2], weight=tiler.weight)
    for tile, coordinates in zip(tiles, tiler.crops):
        # Integrate as batch of size 1
        merger.integrate_batch(tensor_from_rgb_image(tile).unsqueeze(0).float().cuda(), [coordinates])

    merged = merger.merge()
    merged = rgb_image_from_tensor(merged, mean=0, std=1, max_pixel_value=1)
    merged = tiler.crop_to_orignal_size(merged)

    np.testing.assert_equal(merged, image)


def test_tiles_split_merge_2():
    image = np.random.random((5000, 5000, 3)).astype(np.uint8)
    tiler = ImageSlicer(image.shape, tile_size=(512, 512), tile_step=(256, 256), weight="pyramid")

    np.testing.assert_allclose(tiler.weight, tiler.weight.T)

    tiles = tiler.split(image)
    merged = tiler.merge(tiles, dtype=np.uint8)
    np.testing.assert_equal(merged, image)


@skip_if_no_cuda
def test_tiles_split_merge_cuda():
    class MaxChannelIntensity(nn.Module):
        def __init__(self):
            super().__init__()

        def forward(self, input):
            max_channel, _ = torch.max(input, dim=1, keepdim=True)
            return max_channel

    image = np.random.random((5000, 5000, 3)).astype(np.uint8)

    tiler = ImageSlicer(image.shape, tile_size=(512, 512), tile_step=(256, 256), weight="pyramid")
    tiles = [tensor_from_rgb_image(tile) for tile in tiler.split(image)]

    model = MaxChannelIntensity().eval().cuda()

    merger = TileMerger(tiler.target_shape, 1, tiler.weight)
    for tiles_batch, coords_batch in DataLoader(list(zip(tiles, tiler.crops)), batch_size=8, pin_memory=True):
        tiles_batch = tiles_batch.float().cuda()
        pred_batch = model(tiles_batch)

        merger.integrate_batch(pred_batch, coords_batch)

    merged = np.moveaxis(to_numpy(merger.merge()), 0, -1).astype(np.uint8)
    merged = tiler.crop_to_orignal_size(merged)

    np.testing.assert_equal(merged, image.max(axis=2, keepdims=True))



import pytest
import torch

from pytorch_toolbelt.modules.encoders import timm
from pytorch_toolbelt.utils.torch_utils import maybe_cuda, count_parameters


def is_onnx_available():
    try:
        import onnx

        return True
    except ImportError:
        return False


skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available")


@pytest.mark.parametrize(
    ["encoder", "encoder_params"],
    [
        [timm.B0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.MixNetXLEncoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.SKResNet18Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.SWSLResNeXt101Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.TimmResnet200D, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.HRNetW18Encoder, {"pretrained": False}],
        [timm.DPN68Encoder, {"pretrained": False}],
        [timm.NFNetF0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.NFRegNetB0Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
        [timm.TimmRes2Next50Encoder, {"pretrained": False, "layers": [0, 1, 2, 3, 4]}],
    ],
)
@skip_if_no_cuda
def test_jit_trace(encoder, encoder_params):
    model = encoder(**encoder_params).eval()

    print(model.__class__.__name__, count_parameters(model))
    print(model.strides)
    print(model.channels)
    dummy_input = torch.rand((1, 3, 256, 256))
    dummy_input = maybe_cuda(dummy_input)
    model = maybe_cuda(model)

    model = torch.jit.trace(model, dummy_input, check_trace=True)



from typing import Optional
from torch import nn, Tensor
import torch.nn.functional as F
from .functional import label_smoothed_nll_loss

__all__ = ["SoftCrossEntropyLoss"]


class SoftCrossEntropyLoss(nn.Module):
    """
    Drop-in replacement for nn.CrossEntropyLoss with few additions:
    - Support of label smoothing
    """

    __constants__ = ["reduction", "ignore_index", "smooth_factor"]

    def __init__(self, reduction: str = "mean", smooth_factor: float = 0.0, ignore_index: Optional[int] = -100, dim=1):
        super().__init__()
        self.smooth_factor = smooth_factor
        self.ignore_index = ignore_index
        self.reduction = reduction
        self.dim = dim

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        log_prob = F.log_softmax(input, dim=self.dim)
        return label_smoothed_nll_loss(
            log_prob,
            target,
            epsilon=self.smooth_factor,
            ignore_index=self.ignore_index,
            reduction=self.reduction,
            dim=self.dim,
        )



"""Wrappers for different backbones for models that follows Encoder-Decoder architecture.

Encodes listed here provides easy way to swap backbone of classification/segmentation/detection model.
"""

from typing import List

import torch
from torch import Tensor

from .common import EncoderModule, _take, make_n_channel_input
from ..backbone.senet import (
    SENet,
    se_resnext50_32x4d,
    se_resnext101_32x4d,
    se_resnet50,
    se_resnet101,
    se_resnet152,
    senet154,
)

__all__ = [
    "SEResnetEncoder",
    "SEResnet50Encoder",
    "SEResnet101Encoder",
    "SEResnet152Encoder",
    "SEResNeXt50Encoder",
    "SEResNeXt101Encoder",
    "SENet154Encoder",
]


class SEResnetEncoder(EncoderModule):
    """
    The only difference from vanilla ResNet is that it has 'layer0' module
    """

    def __init__(self, seresnet: SENet, channels, strides, layers=None):
        if layers is None:
            layers = [1, 2, 3, 4]
        super().__init__(channels, strides, layers)

        self.maxpool = seresnet.layer0.pool
        del seresnet.layer0.pool

        self.layer0 = seresnet.layer0
        self.layer1 = seresnet.layer1
        self.layer2 = seresnet.layer2
        self.layer3 = seresnet.layer3
        self.layer4 = seresnet.layer4

        self._output_strides = _take(strides, layers)
        self._output_filters = _take(channels, layers)

    @property
    @torch.jit.unused
    def encoder_layers(self):
        return [self.layer0, self.layer1, self.layer2, self.layer3, self.layer4]

    @property
    @torch.jit.unused
    def strides(self):
        return self._output_strides

    @property
    @torch.jit.unused
    def channels(self):
        return self._output_filters

    def forward(self, x: Tensor) -> List[Tensor]:
        output_features = []
        for layer in self.encoder_layers:
            output = layer(x)
            output_features.append(output)

            if layer == self.layer0:
                # Fist maxpool operator is not a part of layer0
                # because we want that layer0 output to have stride of 2
                output = self.maxpool(output)
            x = output

        # Return only features that were requested
        return _take(output_features, self._layers)

    def change_input_channels(self, input_channels: int, mode="auto", **kwargs):
        self.layer0.conv1 = make_n_channel_input(self.layer0.conv1, input_channels, mode, **kwargs)
        return self


class SEResnet50Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = se_resnet50(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class SEResnet101Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = se_resnet101(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class SEResnet152Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = se_resnet152(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class SENet154Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = senet154(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class SEResNeXt50Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = se_resnext50_32x4d(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)


class SEResNeXt101Encoder(SEResnetEncoder):
    def __init__(self, pretrained=True, layers=None):
        encoder = se_resnext101_32x4d(pretrained="imagenet" if pretrained else None)
        super().__init__(encoder, [64, 256, 512, 1024, 2048], [2, 4, 8, 16, 32], layers)



import torch
import pytest

from pytorch_toolbelt.modules.activations import instantiate_activation_block

skip_if_no_cuda = pytest.mark.skipif(not torch.cuda.is_available(), reason="Cuda is not available")


@pytest.mark.parametrize(
    "activation_name",
    ["none", "relu", "relu6", "leaky_relu", "elu", "selu", "celu", "mish", "swish", "hard_sigmoid", "hard_swish"],
)
def test_activations(activation_name):
    act = instantiate_activation_block(activation_name)
    x = torch.randn(128).float()
    y = act(x)
    assert y.dtype == torch.float32


@pytest.mark.parametrize(
    "activation_name",
    ["none", "relu", "relu6", "leaky_relu", "elu", "selu", "celu", "mish", "swish", "hard_sigmoid", "hard_swish"],
)
@skip_if_no_cuda
def test_activations_cuda(activation_name):
    act = instantiate_activation_block(activation_name)
    x = torch.randn(128).float().cuda()
    y = act(x)
    assert y.dtype == torch.float32

    act = instantiate_activation_block(activation_name)
    x = torch.randn(128).half().cuda()
    y = act(x)
    assert y.dtype == torch.float16



from torch import nn

__all__ = ["instantiate_normalization_block", "NORM_BATCH", "NORM_INSTANCE", "NORM_GROUP"]


NORM_BATCH = "batch_norm"
NORM_INSTANCE = "instance_norm"
NORM_GROUP = "group_norm"


def instantiate_normalization_block(normalization: str, in_channels: int, **kwargs):
    if normalization in ("bn", "batch", "batch2d", "batch_norm", "batch_norm_2d", "batchnorm", "batchnorm2d"):
        return nn.BatchNorm2d(num_features=in_channels)

    if normalization in ("bn3d", "batch3d", "batch_norm3d", "batch_norm_3d", "batchnorm3d"):
        return nn.BatchNorm3d(num_features=in_channels)

    if normalization in ("gn", "group", "group_norm", "groupnorm"):
        return nn.GroupNorm(num_channels=in_channels, **kwargs)

    if normalization in (
        "in",
        "instance",
        "instance2d",
        "instance_norm",
        "instancenorm",
        "instance_norm_2d",
        "instancenorm2d",
    ):
        return nn.InstanceNorm2d(num_features=in_channels, **kwargs)

    if normalization in ("in3d", "instance3d", "instance_norm_3d", "instancenorm3d"):
        return nn.InstanceNorm3d(num_features=in_channels, **kwargs)

    raise KeyError(f"Unknown normalization type '{normalization}'")



from typing import Optional
from torch import nn, Tensor
import torch.nn.functional as F
import torch

__all__ = ["FocalCosineLoss"]


class FocalCosineLoss(nn.Module):
    """
    Implementation Focal cosine loss from the "Data-Efficient Deep Learning Method for Image Classification
    Using Data Augmentation, Focal Cosine Loss, and Ensemble" (https://arxiv.org/abs/2007.07805).

    Credit: https://www.kaggle.com/c/cassava-leaf-disease-classification/discussion/203271
    """

    def __init__(self, alpha: float = 1, gamma: float = 2, xent: float = 0.1, reduction="mean"):
        super(FocalCosineLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.xent = xent
        self.reduction = reduction

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        cosine_loss = F.cosine_embedding_loss(
            input,
            torch.nn.functional.one_hot(target, num_classes=input.size(-1)),
            torch.tensor([1], device=target.device),
            reduction=self.reduction,
        )

        cent_loss = F.cross_entropy(F.normalize(input), target, reduction="none")
        pt = torch.exp(-cent_loss)
        focal_loss = self.alpha * (1 - pt) ** self.gamma * cent_loss

        if self.reduction == "mean":
            focal_loss = torch.mean(focal_loss)

        return cosine_loss + self.xent * focal_loss



# coding: utf-8

# Docker names generator, Python port
# https://github.com/shamrin/namesgenerator
# Copyright (c) 2017 Alexey Shamrin
# MIT License

import random

__all__ = ["get_random_name"]

left = [
    "admiring",
    "adoring",
    "affectionate",
    "agitated",
    "amazing",
    "angry",
    "awesome",
    "blissful",
    "boring",
    "brave",
    "clever",
    "cocky",
    "compassionate",
    "competent",
    "condescending",
    "confident",
    "cranky",
    "dazzling",
    "determined",
    "distracted",
    "dreamy",
    "eager",
    "ecstatic",
    "elastic",
    "elated",
    "elegant",
    "eloquent",
    "epic",
    "fervent",
    "festive",
    "flamboyant",
    "focused",
    "friendly",
    "frosty",
    "gallant",
    "gifted",
    "goofy",
    "gracious",
    "happy",
    "hardcore",
    "heuristic",
    "hopeful",
    "hungry",
    "infallible",
    "inspiring",
    "jolly",
    "jovial",
    "keen",
    "kind",
    "laughing",
    "loving",
    "lucid",
    "mystifying",
    "modest",
    "musing",
    "naughty",
    "nervous",
    "nifty",
    "nostalgic",
    "objective",
    "optimistic",
    "peaceful",
    "pedantic",
    "pensive",
    "practical",
    "priceless",
    "quirky",
    "quizzical",
    "relaxed",
    "reverent",
    "romantic",
    "sad",
    "serene",
    "sharp",
    "silly",
    "sleepy",
    "stoic",
    "stupefied",
    "suspicious",
    "tender",
    "thirsty",
    "trusting",
    "unruffled",
    "upbeat",
    "vibrant",
    "vigilant",
    "vigorous",
    "wizardly",
    "wonderful",
    "xenodochial",
    "youthful",
    "zealous",
    "zen",
]

right = [
    # Muhammad ibn Jābir al-Ḥarrānī al-Battānī was a founding father of astronomy.
    # https://en.wikipedia.org/wiki/Mu%E1%B8%A5ammad_ibn_J%C4%81bir_al-%E1%B8%A4arr%C4%81n%C4%AB_al-Batt%C4%81n%C4%AB
    "albattani",
    # Frances E. Allen, became the first female IBM Fellow in 1989. In 2006, she became the first female recipient of the ACM's Turing Award.
    # https://en.wikipedia.org/wiki/Frances_E._Allen
    "allen",
    # June Almeida - Scottish virologist who took the first pictures of the rubella virus - https://en.wikipedia.org/wiki/June_Almeida
    "almeida",
    # Maria Gaetana Agnesi - Italian mathematician, philosopher, theologian and humanitarian.
    # She was the first woman to write a mathematics handbook and the first woman appointed as a
    # Mathematics Professor at a University. https://en.wikipedia.org/wiki/Maria_Gaetana_Agnesi
    "agnesi",
    # Archimedes was a physicist, engineer and mathematician who invented too many things to list them here.
    # https://en.wikipedia.org/wiki/Archimedes
    "archimedes",
    # Maria Ardinghelli - Italian translator, mathematician and physicist - https://en.wikipedia.org/wiki/Maria_Ardinghelli
    "ardinghelli",
    # Aryabhata - Ancient Indian mathematician-astronomer during 476-550 CE https://en.wikipedia.org/wiki/Aryabhata
    "aryabhata",
    # Wanda Austin - Wanda Austin is the President and CEO of The Aerospace Corporation,
    # a leading architect for the US security space programs. https://en.wikipedia.org/wiki/Wanda_Austin
    "austin",
    # Charles Babbage invented the concept of a programmable computer. https://en.wikipedia.org/wiki/Charles_Babbage.
    "babbage",
    # Stefan Banach - Polish mathematician, was one of the founders of modern functional analysis. https://en.wikipedia.org/wiki/Stefan_Banach
    "banach",
    # John Bardeen co-invented the transistor - https://en.wikipedia.org/wiki/John_Bardeen
    "bardeen",
    # Jean Bartik, born Betty Jean Jennings, was one of the original programmers for the ENIAC computer. https://en.wikipedia.org/wiki/Jean_Bartik
    "bartik",
    # Laura Bassi, the world's first female professor https://en.wikipedia.org/wiki/Laura_Bassi
    "bassi",
    # Hugh Beaver, British engineer, founder of the Guinness Book of World Records https://en.wikipedia.org/wiki/Hugh_Beaver
    "beaver",
    # Alexander Graham Bell - an eminent Scottish-born scientist, inventor, engineer and innovator who is
    # credited with inventing the first practical telephone - https://en.wikipedia.org/wiki/Alexander_Graham_Bell
    "bell",
    # Karl Friedrich Benz - a German automobile engineer. Inventor of the first practical motorcar. https://en.wikipedia.org/wiki/Karl_Benz
    "benz",
    # Homi J Bhabha - was an Indian nuclear physicist, founding director, and professor of physics at the Tata Institute of Fundamental Research.
    # Colloquially known as 'father of Indian nuclear programme'- https://en.wikipedia.org/wiki/Homi_J._Bhabha
    "bhabha",
    # Bhaskara II - Ancient Indian mathematician-astronomer whose work on calculus predates Newton and Leibniz
    # by over half a millennium - https://en.wikipedia.org/wiki/Bh%C4%81skara_II#Calculus
    "bhaskara",
    # Elizabeth Blackwell - American doctor and first American woman to receive a medical degree - https://en.wikipedia.org/wiki/Elizabeth_Blackwell
    "blackwell",
    # Niels Bohr is the father of quantum theory. https://en.wikipedia.org/wiki/Niels_Bohr.
    "bohr",
    # Kathleen Booth, she's credited with writing the first assembly language. https://en.wikipedia.org/wiki/Kathleen_Booth
    "booth",
    # Anita Borg - Anita Borg was the founding director of the Institute for Women and Technology (IWT). https://en.wikipedia.org/wiki/Anita_Borg
    "borg",
    # Satyendra Nath Bose - He provided the foundation for Bose–Einstein statistics and the theory of the Bose–Einstein condensate.
    # https://en.wikipedia.org/wiki/Satyendra_Nath_Bose
    "bose",
    # Evelyn Boyd Granville - She was one of the first African-American woman to receive a Ph.D. in mathematics;
    # she earned it in 1949 from Yale University. https://en.wikipedia.org/wiki/Evelyn_Boyd_Granville
    "boyd",
    # Brahmagupta - Ancient Indian mathematician during 598-670 CE who gave rules to compute
    # with zero - https://en.wikipedia.org/wiki/Brahmagupta#Zero
    "brahmagupta",
    # Walter Houser Brattain co-invented the transistor - https://en.wikipedia.org/wiki/Walter_Houser_Brattain
    "brattain",
    # Emmett Brown invented time travel. https://en.wikipedia.org/wiki/Emmett_Brown (thanks Brian Goff)
    "brown",
    # Rachel Carson - American marine biologist and conservationist, her book Silent Spring and other
    # writings are credited with advancing the global environmental movement. https://en.wikipedia.org/wiki/Rachel_Carson
    "carson",
    # Subrahmanyan Chandrasekhar - Astrophysicist known for his mathematical theory on different stages and evolution in structures of the stars.
    # He has won nobel prize for physics - https://en.wikipedia.org/wiki/Subrahmanyan_Chandrasekhar
    "chandrasekhar",
    # Claude Shannon - The father of information theory and founder of digital circuit design theory. (https://en.wikipedia.org/wiki/Claude_Shannon)
    "shannon",
    # Joan Clarke - Bletchley Park code breaker during the Second World War who pioneered techniques
    # that remained top secret for decades. Also an accomplished numismatist https://en.wikipedia.org/wiki/Joan_Clarke
    "clarke",
    # Jane Colden - American botanist widely considered the first female American botanist - https://en.wikipedia.org/wiki/Jane_Colden
    "colden",
    # Gerty Theresa Cori - American biochemist who became the third woman—and first American woman—to win a Nobel Prize in science,
    # and the first woman to be awarded the Nobel Prize in Physiology or Medicine.
    # Cori was born in Prague. https://en.wikipedia.org/wiki/Gerty_Cori
    "cori",
    # Seymour Roger Cray was an American electrical engineer and supercomputer architect who designed a series of computers
    # that were the fastest in the world for decades. https://en.wikipedia.org/wiki/Seymour_Cray
    "cray",
    # This entry reflects a husband and wife team who worked together:
    # Joan Curran was a Welsh scientist who developed radar and invented chaff, a radar countermeasure.
    # https://en.wikipedia.org/wiki/Joan_Curran
    # Samuel Curran was an Irish physicist who worked alongside his wife during WWII and invented the proximity fuse.
    # https://en.wikipedia.org/wiki/Samuel_Curran
    "curran",
    # Marie Curie discovered radioactivity. https://en.wikipedia.org/wiki/Marie_Curie.
    "curie",
    # Charles Darwin established the principles of natural evolution. https://en.wikipedia.org/wiki/Charles_Darwin.
    "darwin",
    # Leonardo Da Vinci invented too many things to list here. https://en.wikipedia.org/wiki/Leonardo_da_Vinci.
    "davinci",
    # Edsger Wybe Dijkstra was a Dutch computer scientist and mathematical scientist. https://en.wikipedia.org/wiki/Edsger_W._Dijkstra.
    "dijkstra",
    # Donna Dubinsky - played an integral role in the development of personal digital assistants (PDAs)
    # serving as CEO of Palm, Inc. and co-founding Handspring. https://en.wikipedia.org/wiki/Donna_Dubinsky
    "dubinsky",
    # Annie Easley - She was a leading member of the team which developed software for the Centaur rocket
    # stage and one of the first African-Americans in her field. https://en.wikipedia.org/wiki/Annie_Easley
    "easley",
    # Thomas Alva Edison, prolific inventor https://en.wikipedia.org/wiki/Thomas_Edison
    "edison",
    # Albert Einstein invented the general theory of relativity. https://en.wikipedia.org/wiki/Albert_Einstein
    "einstein",
    # Gertrude Elion - American biochemist, pharmacologist and the 1988 recipient of
    # the Nobel Prize in Medicine - https://en.wikipedia.org/wiki/Gertrude_Elion
    "elion",
    # Douglas Engelbart gave the mother of all demos: https://en.wikipedia.org/wiki/Douglas_Engelbart
    "engelbart",
    # Euclid invented geometry. https://en.wikipedia.org/wiki/Euclid
    "euclid",
    # Leonhard Euler invented large parts of modern mathematics. https://de.wikipedia.org/wiki/Leonhard_Euler
    "euler",
    # Pierre de Fermat pioneered several aspects of modern mathematics. https://en.wikipedia.org/wiki/Pierre_de_Fermat
    "fermat",
    # Enrico Fermi invented the first nuclear reactor. https://en.wikipedia.org/wiki/Enrico_Fermi.
    "fermi",
    # Richard Feynman was a key contributor to quantum mechanics and particle physics. https://en.wikipedia.org/wiki/Richard_Feynman
    "feynman",
    # Benjamin Franklin is famous for his experiments in electricity and the invention of the lightning rod.
    "franklin",
    # Galileo was a founding father of modern astronomy, and faced politics and obscurantism
    # to establish scientific truth.  https://en.wikipedia.org/wiki/Galileo_Galilei
    "galileo",
    # William Henry 'Bill' Gates III is an American business magnate, philanthropist, investor,
    # computer programmer, and inventor. https://en.wikipedia.org/wiki/Bill_Gates
    "gates",
    # Adele Goldberg, was one of the designers and developers of the Smalltalk language.
    # https://en.wikipedia.org/wiki/Adele_Goldberg_(computer_scientist)
    "goldberg",
    # Adele Goldstine, born Adele Katz, wrote the complete technical description for the
    # first electronic digital computer, ENIAC. https://en.wikipedia.org/wiki/Adele_Goldstine
    "goldstine",
    # Shafi Goldwasser is a computer scientist known for creating theoretical foundations
    # of modern cryptography. Winner of 2012 ACM Turing Award. https://en.wikipedia.org/wiki/Shafi_Goldwasser
    "goldwasser",
    # James Golick, all around gangster.
    "golick",
    # Jane Goodall - British primatologist, ethologist, and anthropologist who is considered
    # to be the world's foremost expert on chimpanzees - https://en.wikipedia.org/wiki/Jane_Goodall
    "goodall",
    # Lois Haibt - American computer scientist, part of the team at IBM that developed FORTRAN - https://en.wikipedia.org/wiki/Lois_Haibt
    "haibt",
    # Margaret Hamilton - Director of the Software Engineering Division of the MIT Instrumentation Laboratory,
    # which developed on-board flight software for the Apollo space program.
    # https://en.wikipedia.org/wiki/Margaret_Hamilton_(scientist)
    "hamilton",
    # Stephen Hawking pioneered the field of cosmology by combining general relativity and
    # quantum mechanics. https://en.wikipedia.org/wiki/Stephen_Hawking
    "hawking",
    # Werner Heisenberg was a founding father of quantum mechanics. https://en.wikipedia.org/wiki/Werner_Heisenberg
    "heisenberg",
    # Grete Hermann was a German philosopher noted for her philosophical work on the foundations of
    # quantum mechanics. https://en.wikipedia.org/wiki/Grete_Hermann
    "hermann",
    # Jaroslav Heyrovský was the inventor of the polarographic method, father of the electroanalytical method, and recipient of the Nobel Prize in 1959.
    # His main field of work was polarography. https://en.wikipedia.org/wiki/Jaroslav_Heyrovsk%C3%BD
    "heyrovsky",
    # Dorothy Hodgkin was a British biochemist, credited with the development of protein crystallography.
    # She was awarded the Nobel Prize in Chemistry in 1964. https://en.wikipedia.org/wiki/Dorothy_Hodgkin
    "hodgkin",
    # Erna Schneider Hoover revolutionized modern communication by inventing a computerized telephone switching method.
    # https://en.wikipedia.org/wiki/Erna_Schneider_Hoover
    "hoover",
    # Grace Hopper developed the first compiler for a computer programming language and  is credited with popularizing
    # the term 'debugging' for fixing computer glitches. https://en.wikipedia.org/wiki/Grace_Hopper
    "hopper",
    # Frances Hugle, she was an American scientist, engineer, and inventor who contributed to the understanding of semiconductors,
    # integrated circuitry, and the unique electrical principles of microscopic materials.
    # https://en.wikipedia.org/wiki/Frances_Hugle
    "hugle",
    # Hypatia - Greek Alexandrine Neoplatonist philosopher in Egypt who was one of the earliest
    # mothers of mathematics - https://en.wikipedia.org/wiki/Hypatia
    "hypatia",
    # Mary Jackson, American mathematician and aerospace engineer who earned the highest title within
    # NASA's engineering department - https://en.wikipedia.org/wiki/Mary_Jackson_(engineer)
    "jackson",
    # Yeong-Sil Jang was a Korean scientist and astronomer during the Joseon Dynasty;
    # he invented the first metal printing press and water gauge. https://en.wikipedia.org/wiki/Jang_Yeong-sil
    "jang",
    # Betty Jennings - one of the original programmers of the ENIAC.
    # https://en.wikipedia.org/wiki/ENIAC - https://en.wikipedia.org/wiki/Jean_Bartik
    "jennings",
    # Mary Lou Jepsen, was the founder and chief technology officer of One Laptop Per Child (OLPC), and the founder of Pixel Qi.
    # https://en.wikipedia.org/wiki/Mary_Lou_Jepsen
    "jepsen",
    # Katherine Coleman Goble Johnson - American physicist and mathematician contributed to the NASA.
    # https://en.wikipedia.org/wiki/Katherine_Johnson
    "johnson",
    # Irène Joliot-Curie - French scientist who was awarded the Nobel Prize for Chemistry in 1935.
    # Daughter of Marie and Pierre Curie. https://en.wikipedia.org/wiki/Ir%C3%A8ne_Joliot-Curie
    "joliot",
    # Karen Spärck Jones came up with the concept of inverse document frequency,
    # which is used in most search engines today. https://en.wikipedia.org/wiki/Karen_Sp%C3%A4rck_Jones
    "jones",
    # A. P. J. Abdul Kalam - is an Indian scientist aka Missile Man of India for his work on the development of
    # ballistic missile and launch vehicle technology - https://en.wikipedia.org/wiki/A._P._J._Abdul_Kalam
    "kalam",
    # Susan Kare, created the icons and many of the interface elements for the original Apple Macintosh in the 1980s,
    # and was an original employee of NeXT, working as the Creative Director.
    # https://en.wikipedia.org/wiki/Susan_Kare
    "kare",
    # Mary Kenneth Keller, Sister Mary Kenneth Keller became the first American woman to earn a PhD in Computer Science in 1965.
    # https://en.wikipedia.org/wiki/Mary_Kenneth_Keller
    "keller",
    # Johannes Kepler, German astronomer known for his three laws of planetary motion -
    # https://en.wikipedia.org/wiki/Johannes_Kepler
    "kepler",
    # Har Gobind Khorana - Indian-American biochemist who shared the 1968 Nobel Prize for Physiology -
    # https://en.wikipedia.org/wiki/Har_Gobind_Khorana
    "khorana",
    # Jack Kilby invented silicone integrated circuits and gave Silicon Valley its name. -
    # https://en.wikipedia.org/wiki/Jack_Kilby
    "kilby",
    # Maria Kirch - German astronomer and first woman to discover a comet -
    # https://en.wikipedia.org/wiki/Maria_Margarethe_Kirch
    "kirch",
    # Donald Knuth - American computer scientist, author of 'The Art of Computer Programming' and creator of the TeX typesetting system.
    # https://en.wikipedia.org/wiki/Donald_Knuth
    "knuth",
    # Sophie Kowalevski - Russian mathematician responsible for important original contributions to analysis,
    # differential equations and mechanics -
    # https://en.wikipedia.org/wiki/Sofia_Kovalevskaya
    "kowalevski",
    # Marie-Jeanne de Lalande - French astronomer, mathematician and cataloguer of stars -
    # https://en.wikipedia.org/wiki/Marie-Jeanne_de_Lalande
    "lalande",
    # Hedy Lamarr - Actress and inventor. The principles of her work are now incorporated into modern Wi-Fi, CDMA and Bluetooth technology.
    # https://en.wikipedia.org/wiki/Hedy_Lamarr
    "lamarr",
    # Leslie B. Lamport - American computer scientist. Lamport is best known for his seminal work in distributed systems and was the winner of the 2013 Turing Award.
    # https://en.wikipedia.org/wiki/Leslie_Lamport
    "lamport",
    # Mary Leakey - British paleoanthropologist who discovered the first fossilized Proconsul skull -
    # https://en.wikipedia.org/wiki/Mary_Leakey
    "leakey",
    # Henrietta Swan Leavitt - she was an American astronomer who discovered the relation between the luminosity and the period of Cepheid variable stars.
    # https://en.wikipedia.org/wiki/Henrietta_Swan_Leavitt
    "leavitt",
    # Daniel Lewin -  Mathematician, Akamai co-founder, soldier, 9/11 victim-- Developed optimization techniques for routing traffic on the internet.
    # Died attempting to stop the 9-11 hijackers.
    # https://en.wikipedia.org/wiki/Daniel_Lewin
    "lewin",
    # Ruth Lichterman - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC - https://en.wikipedia.org/wiki/Ruth_Teitelbaum
    "lichterman",
    # Barbara Liskov - co-developed the Liskov substitution principle. Liskov was also the winner of the Turing Prize in 2008. -
    # https://en.wikipedia.org/wiki/Barbara_Liskov
    "liskov",
    # Ada Lovelace invented the first algorithm. https://en.wikipedia.org/wiki/Ada_Lovelace (thanks James Turnbull)
    "lovelace",
    # Auguste and Louis Lumière - the first filmmakers in history - https://en.wikipedia.org/wiki/Auguste_and_Louis_Lumi%C3%A8re
    "lumiere",
    # Mahavira - Ancient Indian mathematician during 9th century AD who discovered basic algebraic identities -
    # https://en.wikipedia.org/wiki/Mah%C4%81v%C4%ABra_(mathematician)
    "mahavira",
    # Maria Mayer - American theoretical physicist and Nobel laureate in Physics for proposing the nuclear
    # shell model of the atomic nucleus - https://en.wikipedia.org/wiki/Maria_Mayer
    "mayer",
    # John McCarthy invented LISP: https://en.wikipedia.org/wiki/John_McCarthy_(computer_scientist)
    "mccarthy",
    # Barbara McClintock - a distinguished American cytogeneticist, 1983 Nobel Laureate in Physiology or Medicine for discovering transposons.
    # https://en.wikipedia.org/wiki/Barbara_McClintock
    "mcclintock",
    # Malcolm McLean invented the modern shipping container: https://en.wikipedia.org/wiki/Malcom_McLean
    "mclean",
    # Kay McNulty - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Kathleen_Antonelli
    "mcnulty",
    # Lise Meitner - Austrian/Swedish physicist who was involved in the discovery of nuclear fission.
    # The element meitnerium is named after her - https://en.wikipedia.org/wiki/Lise_Meitner
    "meitner",
    # Carla Meninsky, was the game designer and programmer for Atari 2600 games Dodge 'Em and Warlords.
    # https://en.wikipedia.org/wiki/Carla_Meninsky
    "meninsky",
    # Johanna Mestorf - German prehistoric archaeologist and first female museum director in Germany -
    # https://en.wikipedia.org/wiki/Johanna_Mestorf
    "mestorf",
    # Marvin Minsky - Pioneer in Artificial Intelligence, co-founder of the MIT's AI Lab, won the Turing Award in 1969.
    # https://en.wikipedia.org/wiki/Marvin_Minsky
    "minsky",
    # Maryam Mirzakhani - an Iranian mathematician and the first woman to win the Fields Medal.
    # https://en.wikipedia.org/wiki/Maryam_Mirzakhani
    "mirzakhani",
    # Samuel Morse - contributed to the invention of a single-wire telegraph system based on
    # European telegraphs and was a co-developer of the Morse code -
    # https://en.wikipedia.org/wiki/Samuel_Morse
    "morse",
    # Ian Murdock - founder of the Debian project - https://en.wikipedia.org/wiki/Ian_Murdock
    "murdock",
    # John von Neumann - todays computer architectures are based on the von Neumann architecture.
    # https://en.wikipedia.org/wiki/Von_Neumann_architecture
    "neumann",
    # Isaac Newton invented classic mechanics and modern optics. https://en.wikipedia.org/wiki/Isaac_Newton
    "newton",
    # Florence Nightingale, more prominently known as a nurse, was also the first female member of the
    # Royal Statistical Society and a pioneer in statistical graphics
    # https://en.wikipedia.org/wiki/Florence_Nightingale#Statistics_and_sanitary_reform
    "nightingale",
    # Alfred Nobel - a Swedish chemist, engineer, innovator, and armaments manufacturer (inventor of dynamite) -
    # https://en.wikipedia.org/wiki/Alfred_Nobel
    "nobel",
    # Emmy Noether, German mathematician. Noether's Theorem is named after her. https://en.wikipedia.org/wiki/Emmy_Noether
    "noether",
    # Poppy Northcutt. Poppy Northcutt was the first woman to work as part of NASA’s Mission Control.
    # http://www.businessinsider.com/poppy-northcutt-helped-apollo-astronauts-2014-12?op=1
    "northcutt",
    # Robert Noyce invented silicone integrated circuits and gave Silicon Valley its name. -
    # https://en.wikipedia.org/wiki/Robert_Noyce
    "noyce",
    # Panini - Ancient Indian linguist and grammarian from 4th century CE who worked on the world's first formal system -
    # https://en.wikipedia.org/wiki/P%C4%81%E1%B9%87ini#Comparison_with_modern_formal_systems
    "panini",
    # Ambroise Pare invented modern surgery. https://en.wikipedia.org/wiki/Ambroise_Par%C3%A9
    "pare",
    # Louis Pasteur discovered vaccination, fermentation and pasteurization. https://en.wikipedia.org/wiki/Louis_Pasteur.
    "pasteur",
    # Cecilia Payne-Gaposchkin was an astronomer and astrophysicist who, in 1925, proposed
    # in her Ph.D. thesis an explanation for the composition of stars in terms of the relative abundances of hydrogen and helium.
    # https://en.wikipedia.org/wiki/Cecilia_Payne-Gaposchkin
    "payne",
    # Radia Perlman is a software designer and network engineer and most famous for her
    # invention of the spanning-tree protocol (STP). https://en.wikipedia.org/wiki/Radia_Perlman
    "perlman",
    # Rob Pike was a key contributor to Unix, Plan 9, the X graphic system, utf-8, and the Go programming language.
    # https://en.wikipedia.org/wiki/Rob_Pike
    "pike",
    # Henri Poincaré made fundamental contributions in several fields of mathematics.
    # https://en.wikipedia.org/wiki/Henri_Poincar%C3%A9
    "poincare",
    # Laura Poitras is a director and producer whose work, made possible by open source crypto tools, advances
    # the causes of truth and freedom of information by reporting disclosures by
    # whistleblowers such as Edward Snowden. https://en.wikipedia.org/wiki/Laura_Poitras
    "poitras",
    # Claudius Ptolemy - a Greco-Egyptian writer of Alexandria, known as a mathematician, astronomer, geographer, astrologer,
    # and poet of a single epigram in the Greek Anthology - https://en.wikipedia.org/wiki/Ptolemy
    "ptolemy",
    # C. V. Raman - Indian physicist who won the Nobel Prize in 1930 for proposing the Raman effect. -
    # https://en.wikipedia.org/wiki/C._V._Raman
    "raman",
    # Srinivasa Ramanujan - Indian mathematician and autodidact who made extraordinary contributions to
    # mathematical analysis, number theory, infinite series, and continued fractions.
    # https://en.wikipedia.org/wiki/Srinivasa_Ramanujan
    "ramanujan",
    # Sally Kristen Ride was an American physicist and astronaut. She was the first American woman in space,
    # and the youngest American astronaut. https://en.wikipedia.org/wiki/Sally_Ride
    "ride",
    # Rita Levi-Montalcini - Won Nobel Prize in Physiology or Medicine jointly with colleague Stanley Cohen
    # for the discovery of nerve growth factor (https://en.wikipedia.org/wiki/Rita_Levi-Montalcini)
    "montalcini",
    # Dennis Ritchie - co-creator of UNIX and the C programming language. -
    # https://en.wikipedia.org/wiki/Dennis_Ritchie
    "ritchie",
    # Wilhelm Conrad Röntgen - German physicist who was awarded the first Nobel Prize in Physics in 1901
    # for the discovery of X-rays (Röntgen rays). https://en.wikipedia.org/wiki/Wilhelm_R%C3%B6ntgen
    "roentgen",
    # Rosalind Franklin - British biophysicist and X-ray crystallographer whose research was critical to
    # the understanding of DNA - https://en.wikipedia.org/wiki/Rosalind_Franklin
    "rosalind",
    # Meghnad Saha - Indian astrophysicist best known for his development of the Saha equation, used to
    # describe chemical and physical conditions in stars - https://en.wikipedia.org/wiki/Meghnad_Saha
    "saha",
    # Jean E. Sammet developed FORMAC, the first widely used computer language for symbolic manipulation
    # of mathematical formulas. https://en.wikipedia.org/wiki/Jean_E._Sammet
    "sammet",
    # Carol Shaw - Originally an Atari employee, Carol Shaw is said to be the first female video game designer.
    # https://en.wikipedia.org/wiki/Carol_Shaw_(video_game_designer)
    "shaw",
    # Dame Stephanie 'Steve' Shirley - Founded a software company in 1962 employing women working from home.
    # https://en.wikipedia.org/wiki/Steve_Shirley
    "shirley",
    # William Shockley co-invented the transistor - https://en.wikipedia.org/wiki/William_Shockley
    "shockley",
    # Françoise Barré-Sinoussi - French virologist and Nobel Prize Laureate in Physiology or Medicine;
    # her work was fundamental in identifying HIV as the cause of AIDS.
    # https://en.wikipedia.org/wiki/Fran%C3%A7oise_Barr%C3%A9-Sinoussi
    "sinoussi",
    # Betty Snyder - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Betty_Holberton
    "snyder",
    # Frances Spence - one of the original programmers of the ENIAC. https://en.wikipedia.org/wiki/ENIAC -
    # https://en.wikipedia.org/wiki/Frances_Spence
    "spence",
    # Richard Matthew Stallman - the founder of the Free Software movement,
    # the GNU project, the Free Software Foundation, and the League for Programming Freedom.
    # He also invented the concept of copyleft to protect the ideals of this movement, and enshrined
    # this concept in the widely-used GPL (General Public License) for software.
    # https://en.wikiquote.org/wiki/Richard_Stallman
    "stallman",
    # Michael Stonebraker is a database research pioneer and architect of Ingres, Postgres, VoltDB and SciDB.
    # Winner of 2014 ACM Turing Award. https://en.wikipedia.org/wiki/Michael_Stonebraker
    "stonebraker",
    # Janese Swanson (with others) developed the first of the Carmen Sandiego games.
    # She went on to found Girl Tech. https://en.wikipedia.org/wiki/Janese_Swanson
    "swanson",
    # Aaron Swartz was influential in creating RSS, Markdown, Creative Commons, Reddit,
    # and much of the internet as we know it today. He was devoted to freedom of information on the web.
    # https://en.wikiquote.org/wiki/Aaron_Swartz
    "swartz",
    # Bertha Swirles was a theoretical physicist who made a number of contributions to early quantum theory.
    # https://en.wikipedia.org/wiki/Bertha_Swirles
    "swirles",
    # Nikola Tesla invented the AC electric system and every gadget ever used by a James Bond villain.
    # https://en.wikipedia.org/wiki/Nikola_Tesla
    "tesla",
    # Ken Thompson - co-creator of UNIX and the C programming language -
    # https://en.wikipedia.org/wiki/Ken_Thompson
    "thompson",
    # Linus Torvalds invented Linux and Git.
    # https://en.wikipedia.org/wiki/Linus_Torvalds
    "torvalds",
    # Alan Turing was a founding father of computer science.
    # https://en.wikipedia.org/wiki/Alan_Turing.
    "turing",
    # Varahamihira - Ancient Indian mathematician who discovered trigonometric formulae during
    # 505-587 CE - https://en.wikipedia.org/wiki/Var%C4%81hamihira#Contributions
    "varahamihira",
    # Sir Mokshagundam Visvesvaraya - is a notable Indian engineer.
    # He is a recipient of the Indian Republic's highest honour,
    # the Bharat Ratna, in 1955. On his birthday,
    # 15 September is celebrated as Engineer's Day in India in his memory.
    # https://en.wikipedia.org/wiki/Visvesvaraya
    "visvesvaraya",
    # Christiane Nüsslein-Volhard - German biologist, won Nobel Prize in Physiology or Medicine in 1995
    # for research on the genetic control of embryonic development.
    # https://en.wikipedia.org/wiki/Christiane_N%C3%BCsslein-Volhard
    "volhard",
    # Marlyn Wescoff - one of the original programmers of the ENIAC.
    # https://en.wikipedia.org/wiki/ENIAC - https://en.wikipedia.org/wiki/Marlyn_Meltzer
    "wescoff",
    # Andrew Wiles - Notable British mathematician who proved the enigmatic Fermat's Last Theorem -
    # https://en.wikipedia.org/wiki/Andrew_Wiles
    "wiles",
    # Roberta Williams, did pioneering work in graphical adventure games for personal computers,
    # particularly the King's Quest series. https://en.wikipedia.org/wiki/Roberta_Williams
    "williams",
    # Sophie Wilson designed the first Acorn Micro-Computer and the instruction set for ARM processors.
    # https://en.wikipedia.org/wiki/Sophie_Wilson
    "wilson",
    # Jeannette Wing - co-developed the Liskov substitution principle.
    # https://en.wikipedia.org/wiki/Jeannette_Wing
    "wing",
    # Steve Wozniak invented the Apple I and Apple II.
    # https://en.wikipedia.org/wiki/Steve_Wozniak
    "wozniak",
    # The Wright brothers, Orville and Wilbur - credited with inventing and building
    # the world's first successful airplane and making the first controlled, powered and
    # sustained heavier-than-air human flight.
    # https://en.wikipedia.org/wiki/Wright_brothers
    "wright",
    # Rosalyn Sussman Yalow - Rosalyn Sussman Yalow was an American medical physicist,
    # and a co-winner of the 1977 Nobel Prize in Physiology or Medicine for development of the radioimmunoassay technique.
    # https://en.wikipedia.org/wiki/Rosalyn_Sussman_Yalow
    "yalow",
    # Ada Yonath - an Israeli crystallographer,
    # the first woman from the Middle East to win a Nobel prize in the sciences.
    # https://en.wikipedia.org/wiki/Ada_Yonath
    "yonath",
]


def get_random_name(sep="_"):
    r = random.SystemRandom()
    while True:
        name = "%s%s%s" % (r.choice(left), sep, r.choice(right))
        if name == "boring" + sep + "wozniak":  # Steve Wozniak is not boring
            continue
        return name



from typing import Optional, List

import torch.jit
from torch import nn, Tensor

from pytorch_toolbelt.modules import FPNFuse, instantiate_activation_block, instantiate_normalization_block
from pytorch_toolbelt.modules.interfaces import AbstractHead, FeatureMapsSpecification

__all__ = ["HypercolumnHead"]


class HypercolumnHead(AbstractHead):
    """
    Hypercolumn head that concatenates all input feature maps to the size of the largest feature map, compute
    a projection (using conv + bn + act) and resize the projected feature map to the original image size.

    Reference: https://arxiv.org/pdf/1411.5752.pdf
    """

    def __init__(
        self,
        input_spec: FeatureMapsSpecification,
        num_classes: int,
        activation: str,
        normalization: str,
        mid_channels: int,
        output_name: Optional[str] = None,
        dropout_rate: float = 0.0,
        dropout_inplace: bool = False,
        interpolation_mode="bilinear",
        interpolation_align_corners=False,
    ):
        super().__init__(input_spec)
        channels = sum(input_spec.channels)
        self.fuse = FPNFuse(mode=interpolation_mode, align_corners=interpolation_align_corners)

        self.projection = nn.Sequential(
            nn.Conv2d(channels, mid_channels, kernel_size=1),
            instantiate_normalization_block(normalization, mid_channels),
            instantiate_activation_block(activation, inplace=True),
            nn.Dropout2d(dropout_rate, inplace=dropout_inplace),
        )

        self.final = nn.Conv2d(mid_channels, num_classes, kernel_size=3, padding=1)
        self.output_name = output_name
        self.output_spec = FeatureMapsSpecification(channels=(num_classes,), strides=(1,))
        self.interpolation_mode = interpolation_mode
        self.interpolation_align_corners = interpolation_align_corners

    def forward(self, feature_maps: List[Tensor], output_size: torch.Size):
        x = self.fuse(feature_maps)
        x = self.projection(x)
        x = self.final(x)

        output = torch.nn.functional.interpolate(
            x, size=output_size, mode=self.interpolation_mode, align_corners=self.interpolation_align_corners
        )

        if self.output_name is not None:
            return {self.output_name: output}
        else:
            return output



from typing import List

import torch
import torch.nn.functional as F
from torch import nn, Tensor

from ..dsconv import DepthwiseSeparableConv2d

__all__ = ["CANDecoder"]

from ..interfaces import AbstractDecoder, FeatureMapsSpecification


class RCM(nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        self.block = nn.Sequential(
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
        )

    def forward(self, x):
        x = self.conv1(x)
        return self.block(x) + x


def cfm_branch(in_channels: int, out_channels: int, kernel_size: int):
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2, bias=False),
        nn.BatchNorm2d(out_channels),
        nn.ReLU(inplace=True),
        nn.Conv2d(out_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2, bias=False),
        nn.BatchNorm2d(out_channels),
    )


def ds_cfm_branch(in_channels: int, out_channels: int, kernel_size: int):
    return nn.Sequential(
        DepthwiseSeparableConv2d(
            in_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2, bias=False
        ),
        nn.BatchNorm2d(out_channels),
        nn.ReLU(inplace=True),
        DepthwiseSeparableConv2d(
            out_channels, out_channels, kernel_size=kernel_size, padding=kernel_size // 2, bias=False
        ),
        nn.BatchNorm2d(out_channels),
    )


class CFM(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_sizes=[3, 5, 7, 11]):
        super().__init__()
        self.gp_branch = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels),
        )

        self.conv_branches = nn.ModuleList(ds_cfm_branch(in_channels, out_channels, ks) for ks in kernel_sizes)

    def forward(self, x):
        gp = self.gp_branch(x)
        gp = gp.expand_as(x)

        conv_branches = [conv(x) for conv in self.conv_branches]
        return torch.cat(conv_branches + [gp], dim=1)


class AMM(nn.Module):
    def __init__(self, in_channels: int, out_channels: int):
        super().__init__()
        self.conv_bn_relu = nn.Sequential(
            DepthwiseSeparableConv2d(in_channels + out_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, encoder, decoder):
        decoder = F.interpolate(decoder, size=encoder.size()[2:], mode="bilinear", align_corners=False)
        x = torch.cat([encoder, decoder], dim=1)
        x = self.conv_bn_relu(x)
        x = F.adaptive_avg_pool2d(x, 1) * x
        return encoder + x


class CANDecoder(AbstractDecoder):
    """
    Context Aggregation Network
    """

    __constants__ = ["output_spec"]

    def __init__(self, input_spec: FeatureMapsSpecification, out_channels=256):
        super().__init__(input_spec)
        features = input_spec.channels

        self.encoder_rcm = nn.ModuleList(RCM(in_channels, out_channels) for in_channels in features)
        self.cfm = nn.Sequential(CFM(out_channels, out_channels), RCM(out_channels * 5, out_channels))

        self.amm_blocks = nn.ModuleList(AMM(out_channels, out_channels) for in_channels in features[:-1])
        self.rcm_blocks = nn.ModuleList(RCM(out_channels, out_channels) for in_channels in features[:-1])

        self.output_spec = FeatureMapsSpecification(
            channels=[out_channels] * len(input_spec), strides=input_spec.strides
        )

    def forward(self, features: List[Tensor]):
        features = [rcm(x) for x, rcm in zip(features, self.encoder_rcm)]

        x = self.cfm(features[-1])
        outputs = [x]
        num_blocks = len(self.amm_blocks)
        for index in range(num_blocks):
            block_index = num_blocks - index - 1
            encoder_input = features[block_index]
            x = self.amm_blocks[block_index](encoder_input, x)
            x = self.rcm_blocks[block_index](x)
            outputs.append(x)

        return outputs[::-1]

    @torch.jit.unused
    def get_output_spec(self) -> FeatureMapsSpecification:
        return self.output_spec
